/*
 *
 */
#include <LPC177x_8x.h>
#include <uart.h>
#include <crc.h>
#include <debug.h>
#include <stimer.h>
#include <gpio.h>
#include "sdram.h"
#include "version.h"
#include "mpu.h"
#include "dport.h"
#include "dport_hw.h"

static void bootjump(uint32 address);
static void start_main();

extern uint32 * _vect_start;
extern uint32 * _vect_size;

//const struct gpio_t testpin = {
//    LPC_GPIO3, (1 << 30)
//};

/*
 *
 */
int main(void)
{
    int skipflash;

    gpio_init();
    stimer_init(); /* NOTE used by dport, move code to dport itself */
    dport_init();

    uart_setup(DEBUG_UART, UART_BAUD_115200, NULL);

//    rtcmem_get();

    dprint("nsn", "*********************************");
    dprint("sn",  "* "MODULE_NAME);
    dprint("sn",  "* v"VERSION_STRING", "__DATE__" "__TIME__);
    dprint("sn",  "*********************************");

    dprint("<CLK_CCLK = >g4d< MHz>n", 3, CLK_CCLK / 1000 / 1000);
    dprint("<CLK_PCLK = >g4d< MHz>n", 3, CLK_PCLK / 1000 / 1000);
    dprint("<CLK_EMC  = >g4d< MHz>n", 3, CLK_EMC  / 1000 / 1000);

    mpu_init();
    sdram_init();

    if (!sdram_test(SDRAM_TEST_SHORT))
        while (1);

//    if (!sdram_test(gpio_read(&testpin) ? SDRAM_TEST_SHORT : SDRAM_TEST_FULL))
//        while (1);

#if 1
    /*
     * Following code is used for loading main program by dport to FLASH or SDRAM.
     * If RTC general purpose register contains magic number, skip load of main program from flash (if program in SDRAM).
     * "bkpt" instruction is used for entry to dport handler.
     */
    skipflash = 0;
    dprint("s4xn", "MAGIC   ", LPC_RTC->GPREG0);
    
    if (LPC_RTC->GPREG0 == 0x4444BEEF || LPC_RTC->GPREG0 == 0x4444DEAD)
    {
        dprint("<wait debugger>n", 0);
        asm volatile (
            "bkpt #0xAB \n"
            "nop        \n"
            "cpsid i    \n"
            "dsb        \n"
            "isb        \n"
        );
        if (LPC_RTC->GPREG0 == 0x4444DEAD)
            skipflash = 1;
        /* NOTE spoil value to prevent entry to handler each time. */
        LPC_RTC->GPREG0 = 0x00000000;
    }
#endif

    /* check whether we should copy main program to SDRAM */
    if (LOAD_OFFSET >= SDRAM_BASE_ADDR && LOAD_OFFSET < (SDRAM_BASE_ADDR + SDRAM_SIZE))
    {
        /* copy main program to SDRAM */
        if (!skipflash)
        {
            uint8 *rd;
            uint8 *wr;
            uint32 len;
            uint16 crc, fcrc;

            len = *((uint32*)SIZE_OFFSET);
            crc = *((uint16*)CRC_OFFSET);

            dprint("<LENGTH  >4xn", len);

            /* sanity check */
            if (len > SDRAM_SIZE)
            {
                dprint("sn", ERR_PREFIX "size too long");
                for (;;); /* NOTE loop */
            }

            rd = (uint8*)IMAGE_OFFSET;
            wr = (uint8*)LOAD_OFFSET;

            /* check CRC */
            crc16ccitt_init(&fcrc);
            crc16ccitt_update(&fcrc, rd, len);
            dprint("<CRC     >2xn", crc);
            dprint("<FCRC    >2xn", fcrc);
            if (fcrc != crc)
            {
                dprint("sn", ERR_PREFIX "checksum mismatch");
                for (;;); /* NOTE loop */
            }

            while (len--)
                *wr++ = *rd++;
            asm volatile (
                "dsb \n"
                "isb \n"
            );
        }
    }

    /*
     * Copy vector table of main program from LOAD_OFFSET to SRAM.
     *
     * NOTE If LOAD_OFFSET is in flash no neccessary to copy, but
     * this must be performed if LOAD_OFFSET is in SDRAM.
     */
    {
        uint32 len;
        uint8* rd;
        uint8* wr;

        len = (uint32)&_vect_size;
        rd  = (uint8*)LOAD_OFFSET;
        wr  = (uint8*)&_vect_start;

        while (len--)
            *wr++ = *rd++;
        asm volatile (
            "dsb \n"
            "isb \n"
        );
    }

    start_main();

    /* UNREACHED */
    for (;;);

    return 0;
}

/*
 *
 */
static void start_main()
{
    dprint("<LOAD    >4xn", LOAD_OFFSET);

    __disable_irq();
    /*
     * NOTE
     * SDRAM can not be used as vector table (?).
     *
     * In order of loading of main program to SDRAM by dport
     * we should use vector table from SRAM.
     */
    SCB->VTOR = ((uint32)&_vect_start);
    /* NOTE irq stay disabled */

    asm volatile ("dsb");
    asm volatile ("isb");
    bootjump(LOAD_OFFSET);
}

/*
 *
 */
static void __attribute__((naked)) bootjump(uint32 address)
{
    asm volatile (
        "ldr r1, [r0,#0]      \n\t"
        "msr msp, r1          \n\t"
        "ldr PC,[r0,#4]       \n\t"
    );
}

