/*
 * Dumb TCP implementation
 */
#include <LPC177x_8x.h>
#include <os.h>
#include <debug.h>
#include <string.h>
#include <util.h>
#include "net.h"
#include "net_def.h"
#include "arp.h"
#include "tcp.h"

#define DEBUG_TCP

/* NOTE */
#ifndef DEBUG_NET
    #undef DEBUG_TCP
#endif

#ifdef DEBUG_TCP
    #define DPRINT(fmt, ...) dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

static uint16 tcp_cs(struct ipv4_packet_t *ipv4);
static struct tcp_tcb_t tcb;

#define TCP_RX_QUEUE_LEN    4
static struct os_queue_t *qrx;


#define TCP_GET_PACKET(tcp, timeout, length) (os_queue_remove(qrx, OS_FLAG_NONE, \
                OS_MS2TICK(timeout), &tcp, &length) == OS_ERR_NONE)
#define TCP_PUSH_PACKET(tcp, length) os_queue_add(qrx, OS_FLAG_NOWAIT, OS_WAIT_FOREVER, tcp, length)

static inline void _tcb_lock(struct tcp_tcb_t *tcb);
static inline void _tcb_unlock(struct tcp_tcb_t *tcb);

/*
 *
 */
void tcp_init()
{
    qrx = os_queue_init(TCP_RX_QUEUE_LEN, sizeof(struct net_eth_frame_t));
}

/*
 *
 */
static inline void _tcb_lock(struct tcp_tcb_t *tcb)
{
    os_mutex_lock(&tcb->mutex, 1, OS_FLAG_NONE, OS_WAIT_FOREVER);
}

static inline void _tcb_unlock(struct tcp_tcb_t *tcb)
{
    os_mutex_unlock(&tcb->mutex, 1);
}

/*
 * Process incoming TCP packets.
 *
 * ARGS
 *    ipv4    pointer to input ipv4 packet
 *
 * RETURN
 *     count of bytes in response (size of udp packet), 0 if there is no response
 */
void tcp_process(struct ipv4_packet_t *ipv4)
{
    struct tcp_packet_t *tcp;
    int pass;

    DPRINT("<TCP recv:>lbn", NET_HTONS(ipv4->head.total_len), (uint8*)ipv4);

    if (ipv4->head.ihl > 5)
        tcp = (struct tcp_packet_t*)&ipv4->payload[
            (ipv4->head.ihl * 4) - sizeof(struct ipv4_head_t)];
    else
        tcp = (struct tcp_packet_t*)ipv4->payload;

    DPRINT("<doffset  >1xn", tcp->head.doffset );
    DPRINT("<reserved >1xn", tcp->head.reserved);
    DPRINT("<ns       >1xn", tcp->head.ns      );
    DPRINT("<fin      >1xn", tcp->head.fin     );
    DPRINT("<syn      >1xn", tcp->head.syn     );
    DPRINT("<rst      >1xn", tcp->head.rst     );
    DPRINT("<psh      >1xn", tcp->head.psh     );
    DPRINT("<ack      >1xn", tcp->head.ack     );
    DPRINT("<urg      >1xn", tcp->head.urg     );
    DPRINT("<ece      >1xn", tcp->head.ece     );
    DPRINT("<cwr      >1xn", tcp->head.cwr     );

    DPRINT("<checksum >2x<, >2xn", tcp->head.cs, tcp_cs(ipv4));

    /* check for matching connection */
    pass = 0;
    _tcb_lock(&tcb);
    {
        if (tcb.state == TCP_STATE_CLOSED)
        {
            DEBUG_WMSG("state closed");
            goto skip;
        }
        if (tcb.src_ip != ipv4->head.dst_ip || tcb.dst_ip != ipv4->head.src_ip)
        {
            DEBUG_WMSG("ip mismatch");
            goto skip;
        }
        if (tcb.src_port != tcp->head.dst_port || tcb.dst_port != tcp->head.src_port)
        {
            DEBUG_WMSG("port mismatch");
            goto skip;
        }

        pass = 1;
    }
skip:
    _tcb_unlock(&tcb);

    if (pass)
    {
        TCP_PUSH_PACKET(tcp, (NET_HTONS(ipv4->head.total_len) - IPV4_HEAD_LEN(ipv4)));
    }
}

/*
 *
 */
static uint16 tcp_cs(struct ipv4_packet_t *ipv4)
{
    struct tcp_packet_t *tcp;
    uint32 cs;
    int dlen;
    uint16 *data;

    if (ipv4->head.ihl > 5)
        tcp = (struct tcp_packet_t*)&ipv4->payload[
            (ipv4->head.ihl * 4) - sizeof(struct ipv4_head_t)];
    else
        tcp = (struct tcp_packet_t*)ipv4->payload;

    dlen = (NET_HTONS(ipv4->head.total_len) - IPV4_HEAD_LEN(ipv4));

    cs = (ipv4->head.src_ip & 0xffff)   +
         (ipv4->head.src_ip >> 16)      +
         (ipv4->head.dst_ip & 0xffff)   +
         (ipv4->head.dst_ip >> 16)      +
         (ipv4->head.protocol << 8)     +
         NET_HTONS(dlen)                +
         tcp->head.src_port             + 
         tcp->head.dst_port             +
         (tcp->head.seqn & 0xffff)      +
         (tcp->head.seqn >> 16)         +
         (tcp->head.ackn & 0xffff)      +
         (tcp->head.ackn >> 16)         +
         (tcp->head.w3 & 0xffff)        +
         (tcp->head.w3 >> 16 )          +
         tcp->head.urgpointer;

    dlen -= sizeof(struct tcp_packet_head_t);
    data = (uint16*)tcp->payload;

    while (dlen > 1)
    {
        cs += *data++;
        dlen -= 2;
    }
    if (dlen)
        cs += (*data) & 0xff;

    cs = ~(((cs >> 16) + cs) & 0xffff);
//    if (cs == 0)
//        cs = 0xffff;

    return cs;
}

/*
 *
 * RETURN
 *     1 on success, zero otherwise
 */
static int tcp_send(struct tcp_tcb_t *tcb, struct tcp_packet_head_t *head, uint8 *payload, int length)
{
    struct net_eth_frame_t frame; /* NOTE be care of task stack, frame is large enough (1536) */
    struct ipv4_packet_t *ipv4;
    struct tcp_packet_t *tcp;
    uint8 *dst_haddr;
    int arptry;

    /* TODO broadcast packets support */
    arptry = 0;
    dst_haddr = NULL;

    /* get distination MAC */
    while (1)
    {
        dst_haddr = arp_get_haddr(tcb->dst_ip);
        if (dst_haddr == NULL)
        {
            net_send(&frame, arp_request(&frame, NET_HTONL(tcb->dst_ip)));
            arptry++;
            os_wait_ms(50);
        } else {
            break;
        }

        if (arptry >= 2)
        {
            DEBUG_WMSG("ARP retry limit reached");
            return 0;
        }
    }

    ipv4 = (struct ipv4_packet_t *)frame.payload;
    ipv4->head.ihl        = 5;

    /* fill TCP packet */
    tcp = (struct tcp_packet_t*)ipv4_payload(ipv4);
    memcpy(&tcp->head, head, sizeof(struct tcp_packet_head_t));
    tcp->head.src_port = tcb->src_port;
    tcp->head.dst_port = tcb->dst_port;
    memcpy(tcp->payload, payload, length); /* TODO length sanity check */

    length += sizeof(struct tcp_packet_head_t); /* NOTE */
    length += IPV4_HEAD_LEN(ipv4); /* NOTE */

    /* fill IPv4 fields */
    ipv4->head.version      = IPV4_VERSION;
    ipv4->head.dscp         = 0x21;
    ipv4->head.ecn          = 0x00;
    ipv4->head.total_len    = NET_HTONS(length);
    ipv4->head.ident        = NET_HTONS(net.tcp_ident++);
    ipv4->head.f_offset_msb = 0x00;
    ipv4->head.f_offset_lsb = 0x00;
    ipv4->head.flags        = IPV4_FLAG_DF;
    ipv4->head.ttl          = 64;
    ipv4->head.protocol     = IPV4_PROTOCOL_TCP;
    ipv4->head.src_ip       = tcb->src_ip;
    ipv4->head.dst_ip       = tcb->dst_ip;
    ipv4->head.head_cs      = ipv4_head_cs(ipv4);

    /* checksum of TCP calculated from TCP pseudo header, so fill it after IPv4 packet */
    tcp->head.cs = tcp_cs(ipv4);

    /* fill ethernet frame fields */
    memcpy(frame.head.mac_src, net_info.haddr, HADDR_LEN);
    memcpy(frame.head.mac_dst, dst_haddr, HADDR_LEN);
    frame.head.ethertype = NET_HTONS(NET_ETHERTYPE_IPV4);

    DPRINT("<TCP send:>lbn", NET_HTONS(ipv4->head.total_len), (uint8*)ipv4);
    net_send(&frame, (length + sizeof(struct net_eth_frame_head_t)));

    return 1;
}

/*
 *
 * RETURN
 *     pointer to transmission control block or NULL on error 
 */
struct tcp_tcb_t * tcp_open()
{
    struct tcp_packet_t tcp_tx;
    struct tcp_packet_t tcp_rx;
    int dlen;
    int rlen;
    int retry;
    int timeout;

    if (tcb.state != TCP_STATE_CLOSED)
    {
        DEBUG_WMSG("TCP connection not idle");
        return NULL;
    }

    /* NOTE flush receive queue */
    os_queue_flush(qrx);

#if 1
    tcb.dst_ip     = NET_HTONL(net_aton("192.168.1.10"));
    tcb.src_ip     = NET_HTONL(net_aton(NET_IPV4_ADDR));
    tcb.src_port   = NET_HTONS(65100);
    tcb.dst_port   = NET_HTONS(65010);
    tcb.wsize      = TCP_BUF_SIZE;
    tcb.local_mss  = 1000;
#endif

    util_rand_seed(LPC_RTC->SEC + SysTick->CURR);
    tcb.snd_seq = util_rand();

    tcp_tx.head.seqn = tcb.snd_seq;
    tcp_tx.head.ackn = 0;

    dlen = 0;
    {
        uint8 *options;

        options = tcp_tx.options;

        (*options++) = 2; /* kind 2 - maximum segment size */
        (*options++) = 4; /* length */
        (*options++) = tcb.local_mss >> 8;
        (*options++) = tcb.local_mss & 0xff;
        dlen += 4;
    }

    /* send SYN */
    {
        tcp_tx.head.w3  = 0;
        tcp_tx.head.doffset    = (sizeof(struct tcp_packet_head_t) + dlen) / 4;
        tcp_tx.head.syn        = 1;
        tcp_tx.head.wsize      = NET_HTONS(tcb.wsize);
        tcp_tx.head.urgpointer = 0;

        tcb.state = TCP_STATE_SYN_SENT;

#define SYN_SENT_TIMEOUT   500
        retry   = 4; /* XXX */
        timeout = SYN_SENT_TIMEOUT;
        while (retry--)
        {
            if (!tcp_send(&tcb, &tcp_tx.head, tcp_tx.payload, dlen))
            {
                DEBUG_WMSG("send error");
                goto error;
            }

            /* look for ACK */
            if (TCP_GET_PACKET(tcp_rx, timeout, rlen))
            {
//                /* rlen sanity check */
//                if (rlen < sizeof(struct tcp_packet_head_t) + 4 /* MSS */)
//                {
//                    DEBUG_WMSG("tcp packet too short");
//                    goto error;
//                }

                DPRINT("<TCP ACK:>lbn", rlen, (uint8*)&tcp_rx);

                if (tcp_rx.head.rst)
                {
                    DEBUG_WMSG("reset received");
                    goto error;
                }

//                break;
            }

            timeout *= 2; /* NOTE */
        }
    }

    goto error; /* XXX */

    /* TODO MSS sanity check of destination */

    return &tcb;
error:
    tcb.state = TCP_STATE_CLOSED;
    return NULL;
}

