/* 
 *     This file is part of K11, hardware multimedia player.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * OS objects definitions, wrappers, startup routines.
 */
#include <string.h>
#include <LPC177x_8x.h>
#include <os.h>
#include <os_private.h>
#include <port/ARMv7-M/port.h>
#include <debug.h>
#include "osw_objects.h"
#include "gs/gs.h"
#include "dma.h"
#include "player/player.h"
#include "sdcard/sdcard.h"
#include "buttons.h"
#include "vs1053b/decoder.h"
#include "irqp.h"
#include "gpioirq.h"
#include "suspend.h"
#include "net/net.h"

struct osw_task_info_t {
    char name[16];
    char enabled;       /* 1 - initialize task, 0 - not initialize */
    BASE_TYPE ssize;    /* size of stack */
    BASE_TYPE priority; /* NOTE not implemented */
    void (*entry)();
    void *context;
};


static const struct osw_task_info_t tinfo[OS_CONFIG_TASK_COUNT] = {
    /* task's name    enabled    stack size            priority,     entry point,       context */
    {"GS       ",      1,               256,             3,    gs_task,           NULL},
    {"Player   ",      1,          4 * 1024,             1,    player_task,       NULL},
    {"Decoder  ",      1,          4 * 1024,             0,    decoder_task,      NULL},
    {"Buttons  ",      1,          1 * 1024,           255,    buttons_task,      NULL},
    {"Net      ",      1,          8 * 1024,             1,    net_task,          NULL},
    {"Net Proc ",      1,          8 * 1024,             2,    net_process_task,  NULL},
    {"Net Test ",      0,          8 * 1024,             3,    net_test_task,     NULL},
    {"Suspend  ",      1,          2 * 1024,           255,    suspend_task,      NULL},
};

//#define OSW_STACK_SPOOL_SIZE    (DEFAULT_STACK_SIZE * OS_CONFIG_TASK_COUNT)
#define OSW_STACK_SPOOL_SIZE    (50 * 1024)
uint8 sspool[OSW_STACK_SPOOL_SIZE] __attribute__((section("sram")));

void osw_print_tasks_state();
void osw_check_tasks_stack();
void osw_void_handler();

extern uint32 *_uvect_start;
extern uint32 *_uvect_size;

void * userfunctions[] = {
    &osw_print_tasks_state, /* 0 */
    &osw_check_tasks_stack, /* 1 */
    &osw_void_handler,      /* 2 */
    &osw_void_handler,      /* 3 */
};

/*
 * Initialize tasks, queues and other OS related structures.
 */
void osw_init()
{
    int i;
    BASE_TYPE sstart;

    /* Copy user-defined table of functions */
    {
        void *table;
        uint32 size;

        table =         &_uvect_start;
        size  = (uint32)&_uvect_size;

        memcpy(table, userfunctions, size);
    }

#define STACK_MARK 0xAB
    /* NOTE for debug purposes fill stack with initial values */
    memset(&sspool, STACK_MARK, OSW_STACK_SPOOL_SIZE);

    /*
     * Initialize tasks
     */
    dprint("sn", "Init tasks:");
    sstart = 0;
    for (i = 0; i < OS_CONFIG_TASK_COUNT; i++)
    {
        struct osw_task_info_t *ti;

        ti = (struct osw_task_info_t *)&tinfo[i];
        if ((sstart + ti->ssize) > OSW_STACK_SPOOL_SIZE)
        {
            dprint("sn", ERR_PREFIX "stack spool overflow");
            break;
        }

        {
            BASE_TYPE saddr_start;
            BASE_TYPE saddr_end;
            BASE_TYPE entry;

            saddr_start = (BASE_TYPE)sspool + sstart;
            saddr_end   = saddr_start + ti->ssize - 1;
            entry       = (BASE_TYPE)ti->entry;

            dprint("tg1d_ c_s",
                    2, i, 
                    ti->enabled ? '+' : '-',
                    ti->name);

            if (ti->enabled && ti->entry != NULL)
            {
                dprint(" 4x*:4xs",
                        saddr_start,
                        saddr_end, ", ");
                if (ti->entry)
                    dprint("4xn", entry);
                else
                    dprint("sn", "NULL");

                OS_TASK_INIT(ti->name, &sspool[sstart], ti->ssize, ti->priority, ti->entry, ti->context);
                sstart += ti->ssize;
            } else {
                dprint("sn", "*");
            }
        }
    }

    /*
     * Set interrupts priority levels to minimum.
     * Priorities of interrupts will be changed in apprioriate tasks.
     */
    {
        NVIC_SetPriorityRaw(WDT_IRQn          , LOWEST_IRQP);  /*!< Watchdog Timer Interrupt                         */
        NVIC_SetPriorityRaw(TIMER0_IRQn       , LOWEST_IRQP);  /*!< Timer0 Interrupt                                 */
        NVIC_SetPriorityRaw(TIMER1_IRQn       , LOWEST_IRQP);  /*!< Timer1 Interrupt                                 */
        NVIC_SetPriorityRaw(TIMER2_IRQn       , LOWEST_IRQP);  /*!< Timer2 Interrupt                                 */
        NVIC_SetPriorityRaw(TIMER3_IRQn       , LOWEST_IRQP);  /*!< Timer3 Interrupt                                 */
#if 0
        /* NOTE DPort */
        NVIC_SetPriorityRaw(UART0_IRQn        , LOWEST_IRQP);  /*!< UART0 Interrupt                                  */
#endif
        NVIC_SetPriorityRaw(UART1_IRQn        , LOWEST_IRQP);  /*!< UART1 Interrupt                                  */
        NVIC_SetPriorityRaw(UART2_IRQn        , LOWEST_IRQP);  /*!< UART2 Interrupt                                  */
        NVIC_SetPriorityRaw(UART3_IRQn        , LOWEST_IRQP);  /*!< UART3 Interrupt                                  */
        NVIC_SetPriorityRaw(PWM1_IRQn         , LOWEST_IRQP);  /*!< PWM1 Interrupt                                   */
        NVIC_SetPriorityRaw(I2C0_IRQn         , LOWEST_IRQP);  /*!< I2C0 Interrupt                                   */
        NVIC_SetPriorityRaw(I2C1_IRQn         , LOWEST_IRQP);  /*!< I2C1 Interrupt                                   */
        NVIC_SetPriorityRaw(I2C2_IRQn         , LOWEST_IRQP);  /*!< I2C2 Interrupt                                   */
        NVIC_SetPriorityRaw(SSP0_IRQn         , LOWEST_IRQP);  /*!< SSP0 Interrupt                                   */
        NVIC_SetPriorityRaw(SSP1_IRQn         , LOWEST_IRQP);  /*!< SSP1 Interrupt                                   */
        NVIC_SetPriorityRaw(PLL0_IRQn         , LOWEST_IRQP);  /*!< PLL0 Lock (Main PLL) Interrupt                   */
        NVIC_SetPriorityRaw(RTC_IRQn          , LOWEST_IRQP);  /*!< Real Time Clock Interrupt                        */
        NVIC_SetPriorityRaw(EINT0_IRQn        , LOWEST_IRQP);  /*!< External Interrupt 0 Interrupt                   */
        NVIC_SetPriorityRaw(EINT1_IRQn        , LOWEST_IRQP);  /*!< External Interrupt 1 Interrupt                   */
        NVIC_SetPriorityRaw(EINT2_IRQn        , LOWEST_IRQP);  /*!< External Interrupt 2 Interrupt                   */
        NVIC_SetPriorityRaw(EINT3_IRQn        , LOWEST_IRQP);  /*!< External Interrupt 3 Interrupt                   */
        NVIC_SetPriorityRaw(ADC_IRQn          , LOWEST_IRQP);  /*!< A/D Converter Interrupt                          */
        NVIC_SetPriorityRaw(BOD_IRQn          , LOWEST_IRQP);  /*!< Brown-Out Detect Interrupt                       */
        NVIC_SetPriorityRaw(USB_IRQn          , LOWEST_IRQP);  /*!< USB Interrupt                                    */
        NVIC_SetPriorityRaw(CAN_IRQn          , LOWEST_IRQP);  /*!< CAN Interrupt                                    */
        NVIC_SetPriorityRaw(DMA_IRQn          , LOWEST_IRQP);  /*!< General Purpose DMA Interrupt                    */
        NVIC_SetPriorityRaw(I2S_IRQn          , LOWEST_IRQP);  /*!< I2S Interrupt                                    */
        NVIC_SetPriorityRaw(ENET_IRQn         , LOWEST_IRQP);  /*!< Ethernet Interrupt                               */
        NVIC_SetPriorityRaw(MCI_IRQn          , LOWEST_IRQP);  /*!< SD/MMC card I/F Interrupt                        */
        NVIC_SetPriorityRaw(MCPWM_IRQn        , LOWEST_IRQP);  /*!< Motor Control PWM Interrupt                      */
        NVIC_SetPriorityRaw(QEI_IRQn          , LOWEST_IRQP);  /*!< Quadrature Encoder Interface Interrupt           */
        NVIC_SetPriorityRaw(PLL1_IRQn         , LOWEST_IRQP);  /*!< PLL1 Lock (USB PLL) Interrupt                    */
        NVIC_SetPriorityRaw(USBActivity_IRQn  , LOWEST_IRQP);  /*!< USB Activity interrupt                           */
        NVIC_SetPriorityRaw(CANActivity_IRQn  , LOWEST_IRQP);  /*!< CAN Activity interrupt                           */
        NVIC_SetPriorityRaw(UART4_IRQn        , LOWEST_IRQP);  /*!< UART4 Interrupt                                  */
        NVIC_SetPriorityRaw(SSP2_IRQn         , LOWEST_IRQP);  /*!< SSP2 Interrupt                                   */
        NVIC_SetPriorityRaw(LCD_IRQn          , LOWEST_IRQP);  /*!< LCD Interrupt                                    */
        NVIC_SetPriorityRaw(GPIO_IRQn         , LOWEST_IRQP);  /*!< GPIO Interrupt                                   */
        NVIC_SetPriorityRaw(PWM0_IRQn         , LOWEST_IRQP);  /*!< PWM0 Interrupt                                   */
        NVIC_SetPriorityRaw(EEPROM_IRQn       , LOWEST_IRQP);  /*!< EEPROM Interrupt                           */
    }

    /* NOTE initialize DMA */
    dma_init();
}

/*
 * print tasks state
 */
static void osw_print_lock_state(struct os_task_lock_t *lock);

void osw_print_tasks_state()
{
    int i;
    volatile struct os_taskcb_t *ptcb;
    struct os_task_context_t    *tcont;

    ptcb = os_tasks;
    for (i = 0; i < os_taskidx; i++)
    {
#ifdef OS_CONFIG_TASK_NAME_SIZE
        dprint("<Task #>4d_*(s*)sn", i, ptcb->name, ptcb == os_current_taskcb ? " (CURRENT)" : "");
#else
        dprint("<Task #>4dn", i);
#endif
        dprint("t<TOS   >4xn", ptcb->tos);
        tcont = (struct os_task_context_t *)ptcb->tos;
        dprint("tsn", "-context-");
        dprint("tt<r0    >4xn", tcont->r0);
        dprint("tt<r1    >4xn", tcont->r1);
        dprint("tt<r2    >4xn", tcont->r2);
        dprint("tt<r3    >4xn", tcont->r3);
        dprint("tt<r4    >4xn", tcont->r4);
        dprint("tt<r5    >4xn", tcont->r5);
        dprint("tt<r6    >4xn", tcont->r6);
        dprint("tt<r7    >4xn", tcont->r7);
        dprint("tt<r8    >4xn", tcont->r8);
        dprint("tt<r9    >4xn", tcont->r9);
        dprint("tt<r10   >4xn", tcont->r10);
        dprint("tt<r11   >4xn", tcont->r11);
        dprint("tt<r12   >4xn", tcont->r12);
        dprint("tt<lr    >4xn", tcont->lr);
        dprint("tt<pc    >4xn", tcont->pc);
        dprint("tt<xpsr  >4xn", tcont->xpsr);
        if (ptcb->lock.state & OS_TASK_LOCKED)
            osw_print_lock_state((struct os_task_lock_t *)&ptcb->lock);
        dprint("tsn", "---");
#ifdef OS_USE_TIMEOUT
        dprint("tt<timeout  >4xsn", ptcb->timeout, ptcb->timeout ? "" : " (FOREVER)");
#endif
#ifdef OS_CONFIG_USE_TASK_SLICE
        dprint("tt<tslice   >4xn", ptcb->tslice);
#ifdef OS_CONFIG_USE_VARIABLE_TASK_SLICE
        dprint("tt<dslice   >4xn", ptcb->dslice);
#endif
#endif
        ptcb++;
    }
}

static void osw_print_lock_state(struct os_task_lock_t *lock)
{
    char *lname;
    dprint("tsn", "-lock-");

    switch (lock->state)
    {
        case OS_TASK_STATE_RUN               : lname = "RUN";                break;
        case OS_TASK_STATE_SUSPEND           : lname = "SUSPEND";            break;
#if OS_USE_LOCK
        case OS_TASK_STATE_LOCKED_WAIT       : lname = "LOCKED WAIT";        break;
        case OS_TASK_STATE_LOCKED_MUTEX      : lname = "LOCKED MUTEX";       break;
        case OS_TASK_STATE_LOCKED_EVENT      : lname = "LOCKED EVENT";       break;
        case OS_TASK_STATE_LOCKED_QUEUE_FULL : lname = "LOCKED QUEUE FULL";  break;
        case OS_TASK_STATE_LOCKED_QUEUE_EMPTY: lname = "LOCKED QUEUE EMPTY"; break;
        case OS_TASK_STATE_LOCKED_MULTI      : lname = "LOCKED MULTI";       break;
#endif
        default:
            lname = "";
    }
    dprint("tt<state    >4x_*(s*)n", lock->state, lname);

    switch (lock->state)
    {
#if OS_USE_LOCK
        case OS_TASK_STATE_LOCKED_WAIT:
            break;
#ifdef OS_CONFIG_USE_MUTEX
        case OS_TASK_STATE_LOCKED_MUTEX:
            dprint("tt<pobj     >4xn",  lock->pobj);
            dprint("tt<mask     >4xn",  lock->mask);
            break;
#endif
#ifdef OS_CONFIG_USE_EVENT
        case OS_TASK_STATE_LOCKED_EVENT:
            dprint("tt<pobj     >4xn",  lock->pobj);
            dprint("tt<mask     >4xn",  lock->mask);
            break;
#endif
#ifdef OS_CONFIG_USE_QUEUE
        case OS_TASK_STATE_LOCKED_QUEUE_FULL:
            {
                struct os_queue_t *q;
                q = (struct os_queue_t*)lock->pobj;

                dprint("tt<pobj     >4xn",  lock->pobj);
                dprint("ttsn","-queue-");
                dprint("ttt<mutex     >4xn",  q->mutex);
                dprint("ttt<qsize     >4xn",  q->qsize);
                dprint("ttt<msize     >4xn",  q->msize);
                dprint("ttt<count     >4xn",  q->count);
                dprint("ttt<wp        >4xn",  q->wp);
                dprint("ttt<spool     >4xn",  q->spool);
            }
            break;
        case OS_TASK_STATE_LOCKED_QUEUE_EMPTY:
            {
                struct os_queue_t *q;
                q = (struct os_queue_t*)lock->pobj;

                dprint("tt<pobj     >4xn",  lock->pobj);
                dprint("ttsn","-queue-");
                dprint("ttt<mutex     >4xn",  q->mutex);
                dprint("ttt<qsize     >4xn",  q->qsize);
                dprint("ttt<msize     >4xn",  q->msize);
                dprint("ttt<count     >4xn",  q->count);
                dprint("ttt<wp        >4xn",  q->wp);
                dprint("ttt<spool     >4xn",  q->spool);
            }
            break;
#endif
#ifdef OS_CONFIG_USE_MULTI
        case OS_TASK_STATE_LOCKED_MULTI:
            {
                struct os_multi_event_t *multi;
                struct os_task_lock_t *event;
                int n;

                multi = (struct os_multi_event_t*)lock->pobj;

                dprint("tt<pobj     >4xn",  lock->pobj);
                dprint("ttsn","-multi-");
                dprint("ttt<count     >4xn",  multi->count);
                dprint("ttt<n         >4xn",  multi->n);
                dprint("ttt<type      >4xn",  multi->type);
                dprint("ttt<events    >4xn",  multi->events);
                dprint("ttsn","---");
                n     = multi->n;
                event = multi->events;
                while (n--)
                {
                    /* NOTE call recursively */
                    osw_print_lock_state(event++);
                }
            }
            break;
#endif
#endif /* OS_USE_LOCK */
    }
}


/*
 * Print stack water level
 *
 * XXX TODO unify with task_init
 */
void osw_check_tasks_stack()
{
    int i;
    BASE_TYPE sstart;

    /*
     * Initialize tasks
     */
    dprint("sn", "Water level:");
    sstart = 0;
    for (i = 0; i < OS_CONFIG_TASK_COUNT; i++)
    {
        struct osw_task_info_t *ti;

        ti = (struct osw_task_info_t *)&tinfo[i];
        if ((sstart + ti->ssize) > OSW_STACK_SPOOL_SIZE)
            break;

        {
            BASE_TYPE saddr_start;
            BASE_TYPE saddr_end;

            saddr_start = (BASE_TYPE)sspool + sstart;
            saddr_end   = saddr_start + ti->ssize - 1;

            dprint("tg1d_ c_s",
                    2, i, 
                    ti->enabled ? '+' : '-',
                    ti->name);

            if (ti->enabled && ti->entry != NULL)
            {
                int level;
                uint8 *mark;

                dprint(" 4x*:4xs",
                        saddr_start,
                        saddr_end, ", ");

                level = saddr_start;
                mark = &sspool[sstart];
                while (level < saddr_end && *mark == STACK_MARK)
                {
                    level++;
                    mark++;
                }
                dprint("<level >4x<, used >g4d<, avail >4dn",
                        level,
                        6,
                        saddr_end + 1 - level,
                        level - saddr_start);

                sstart += ti->ssize;
            } else {
                dprint("sn", "");
            }
        }
    }
}

/*
 *
 */
void osw_void_handler()
{


}

