/* 
 *     This file is part of K11, hardware multimedia player.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <string.h>
#include <os.h>
#include <debug.h>
#include <stimer.h>
#include <util.h>
#include "vs1053b.h"
#include "decoder.h"
#include "../player/player.h"

#define DEBUG_DECODER

#ifdef DEBUG_DECODER
    #define DPRINT(fmt, ...)  dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

enum {
    DECODER_FEED_END,
    DECODER_FEED_STOPPED
};
static int decoder_feed();
static int decoder_resume();

enum send_position_type_t {
    DECODER_SEND_POSITION_NONE,
    DECODER_SEND_POSITION_NORMAL,
    DECODER_SEND_POSITION_FORCED,
};
static void decoder_send_position(enum send_position_type_t type);

#define DECODER_PROCMSG_NOP     0
#define DECODER_PROCMSG_STOP    1
static int decoder_process_msg();

#ifdef DEC_TEST_GPIO
const struct gpio_t dec_testpins[2] = {
    {LPC_GPIO3, (1 << 28)}, /* MP5, P3[28]  */
    {LPC_GPIO3, (1 << 29)}, /* MP11, P3[29] */
};

#endif

struct decoder_t {
    uint32 flen;    /* whole file length */
    uint32 fpos;    /* current decode position */

    uint32 res_fpos;  /* resume position */
    uint32 res_time;  /* resume time */
    uint32 res_pause; /* resume paused state */

    uint32 posto;   /* send position timer */

#define DECODER_CMD_PAUSE    (1 << 0) /* pause playback */
    uint32 cmd;
};


static struct decoder_t decoder;

/*
 *
 */
void decoder_task()
{
    struct decoder_msg_t dmsg;
    struct player_msg_t pmsg;
    BASE_TYPE msglen;
    int play, resume;

    os_event_wait(&player.event, PLAYER_EVENT_INIT, OS_FLAG_NONE, OS_WAIT_FOREVER);
//    vs1053b_set_volume(30);

#ifdef DEC_TEST_GPIO
    gpio_setdir(DECTEST_GPIO0, GPIO_DIR_OUTPUT);
    gpio_setdir(DECTEST_GPIO1, GPIO_DIR_OUTPUT);

    gpio_drive(DECTEST_GPIO0, 0);
    gpio_drive(DECTEST_GPIO1, 0);
#endif

    while (1)
    {
        /* cleanup all commands */
        decoder.cmd = 0;

        vs1053b_set_low_fclk();

        play   = 0;
        resume = 0;
        do
        {
            if (os_queue_remove(player.qdecoder, OS_FLAG_NONE, OS_WAIT_FOREVER,
                        &dmsg, &msglen) == OS_ERR_NONE)
            {
                switch (dmsg.id)
                {
                    case DECODER_MSG_ID_RESUME:
                        decoder.flen      = dmsg.resume.flen;
                        decoder.res_fpos  = dmsg.resume.fpos;
                        decoder.res_time  = dmsg.resume.time;
                        decoder.res_pause = dmsg.resume.pause;

                        resume = 1;
                        break;
                    case DECODER_MSG_ID_PLAY:
                        decoder.flen  = dmsg.play.flen;
                        decoder.fpos  = 0;
                        play = 1;

                        /* send message that we are started playback */
                        pmsg.id = PLAYER_MSG_ID_DEC_PLAY;
                        PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_empty_t);

                        break;
                    case DECODER_MSG_ID_STOP:
                        /* send confirm that we are idle and player can initialize playback */
                        pmsg.id = PLAYER_MSG_ID_DEC_STOPPED;
                        PLAYER_SEND_MSG(&pmsg, pmsg_empty_t);
                        break;
                    case DECODER_MSG_ID_VOLUME:
                        {
                            int value;
                            value = decoder_set_volume(dmsg.volume.value);

                            /* send real volume information */
                            pmsg.id           = PLAYER_MSG_ID_DEC_VOLUME;
                            pmsg.volume.value = value;
                            PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_volume_t);
                        }
                        break;
                    default:
                        DEBUG_WMSGF("0. odd message", "1xn", dmsg.id);
                } 
            }
        } while (!(play || resume));

        vs1053b_set_decode_time(0);

        if (resume)
        {
            if (!decoder_resume())
            {
                pmsg.id = PLAYER_MSG_ID_DEC_STOPPED;
                PLAYER_SEND_MSG(&pmsg, pmsg_empty_t);
                goto cancel;
            }
        }

        stimer_settime(&decoder.posto);

        /* 
         * query next track or send stopped message
         */
        if (decoder_feed() == DECODER_FEED_STOPPED)
        {
            pmsg.id = PLAYER_MSG_ID_DEC_STOPPED;
            PLAYER_SEND_MSG(&pmsg, pmsg_empty_t);
        } else {
            /* NOTE send in blocking mode */
            pmsg.id = PLAYER_MSG_ID_DEC_NEXTTRACK;
            PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_empty_t);
        }

cancel:
        vs1053b_cancel();
    }
}

/*
 * send position information to player
 */
static void
decoder_send_position(
        enum send_position_type_t type)
{
    struct player_msg_t pmsg;

#define DECODER_NORMAL_SEND_POSITION_TO    1000
#define DECODER_SEND_POSITION_TO           DECODER_NORMAL_SEND_POSITION_TO

    if (type == DECODER_SEND_POSITION_NONE)
    {
        pmsg.id = PLAYER_MSG_ID_DEC_POSITION;
        pmsg.position.time  = 0;
        pmsg.position.value = 0;
        PLAYER_SEND_MSG(&pmsg, pmsg_position_t);
    } else if (
            stimer_deltatime(decoder.posto) >= DECODER_SEND_POSITION_TO ||
            type == DECODER_SEND_POSITION_FORCED
    ) {
        stimer_settime(&decoder.posto);
        pmsg.id = PLAYER_MSG_ID_DEC_POSITION;

        vs1053b_set_low_fclk();
        pmsg.position.time = vs1053b_get_decode_time();
        vs1053b_set_hi_fclk();

        pmsg.position.value = decoder.fpos * 100 / decoder.flen;
        pmsg.position.fpos  = decoder.fpos;
        PLAYER_SEND_MSG(&pmsg, pmsg_position_t);
    }
}

/*
 *
 */
int decoder_set_volume(int value)
{
    int ret;

    vs1053b_set_low_fclk();
    ret = vs1053b_set_volume(value);
    vs1053b_set_hi_fclk();

    return ret;
}

/*
 *
 */
int decoder_set_bass(int bass, int dir, int value)
{
    int ret;
    vs1053b_set_low_fclk();
    ret = vs1053b_set_bass(bass, dir, value);
    vs1053b_set_hi_fclk();

    return ret;
}


/*
 * feed vs1053 with data
 *
 * RETURN
 *    1 if playback was stopped, 0 if full file was exhausted
 */
static int decoder_feed()
{
    struct fcache_entry_t *entry;
    uint8 *buf;
    int n;
    int drained;
    int ret;

    /* precache */
    player_fcache_fill();

    /* process messages from player (for early pause on resume) */
    if (decoder_process_msg() == DECODER_PROCMSG_STOP)
    {
        ret = DECODER_FEED_STOPPED;
        goto out;
    }

    /* data to vs1053 can be feeded on higher clock rate */
    vs1053b_set_hi_fclk();

    ret = DECODER_FEED_END;
    while (1)
    {
        entry = player_fcache_get(&drained);
        if (!entry)
        {
            /*
             * If file was drained and cache underflow was occured we can stop
             * decoding.
             */
            if (drained)
            {
                ret = DECODER_FEED_END;
                goto out;
            }

            DEBUG_WMSG("cache underflow");

            /* fill cache */
            player_fcache_fill();

            /* process messages from player */
            if (decoder_process_msg() == DECODER_PROCMSG_STOP)
            {
                ret = DECODER_FEED_STOPPED;
                goto out;
            }

            continue;
        }

        buf = entry->buf;

#define FEED_PORTION    32
        while (entry->len)
        {
            n = entry->len;
            if (n > FEED_PORTION)
                n = FEED_PORTION;

#ifdef DEC_TEST_GPIO
            gpio_drive(DECTEST_GPIO0, 1);
#endif
            vs1053b_writesdi(buf, n);
#ifdef DEC_TEST_GPIO
            gpio_drive(DECTEST_GPIO0, 0);
#endif

            buf          += n;
            entry->len   -= n;
            decoder.fpos += n;

            if (!vs1053b_check_dreq())
            {
                /* fill cache */
                player_fcache_fill();

                /* send position information to player */
                decoder_send_position(DECODER_SEND_POSITION_NORMAL);

                /* process messages from player */
                if (decoder_process_msg() == DECODER_PROCMSG_STOP)
                {
                    ret = DECODER_FEED_STOPPED;
                    goto out;
                }

#ifdef DEC_TEST_GPIO
                gpio_drive(DECTEST_GPIO1, 1);
#endif
                vs1053b_wait_dreq();
#ifdef DEC_TEST_GPIO
                gpio_drive(DECTEST_GPIO1, 0);
#endif
            }
        }
        player_fcache_remove();
    }

out:
    decoder_send_position(DECODER_SEND_POSITION_NONE);
    return ret;
}


/*
 * Return 1 if decoder should be stopped
 */
static int decoder_process_msg()
{
    struct player_msg_t pmsg;
    struct decoder_msg_t dmsg;
    BASE_TYPE qres;

#define MESSAGE_CHECK_TO         500    /* ms */
#define PLAYBACK_INTERRUPT_TO   1000    /* ms */

    while (1)
    {
        if (decoder.cmd & DECODER_CMD_PAUSE)
            qres = os_queue_remove(player.qdecoder, OS_FLAG_NONE, OS_MS2TICK(MESSAGE_CHECK_TO), &dmsg, NULL);
        else
            qres = os_queue_remove(player.qdecoder, OS_FLAG_NOWAIT, OS_WAIT_FOREVER, &dmsg, NULL);
        if (qres == OS_ERR_NONE)
        {
            switch (dmsg.id)
            {
                case DECODER_MSG_ID_STOP:
                    dprint("sn", "stop");
                    return DECODER_PROCMSG_STOP; /* NOTE */
                case DECODER_MSG_ID_PAUSE:
                    if (decoder.cmd & DECODER_CMD_PAUSE)
                        BITMASK_CLEAR(decoder.cmd, DECODER_CMD_PAUSE);
                    else
                        BITMASK_SET(decoder.cmd, DECODER_CMD_PAUSE);
                    break;
                case DECODER_MSG_ID_VOLUME:
                    {
                        int value;
                        value = decoder_set_volume(dmsg.volume.value);

                        /* send real volume information */
                        pmsg.id           = PLAYER_MSG_ID_DEC_VOLUME;
                        pmsg.volume.value = value;
                        PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_volume_t);
                    }
                    break;
                case DECODER_MSG_ID_BASS:
                    {
                        int value;
                        value = decoder_set_bass(dmsg.bass.bass, dmsg.bass.dir, 0);

                        /* send real bass information */
                        pmsg.id           = PLAYER_MSG_ID_DEC_BASS;
                        pmsg.bass.value   = value;
                        PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_bass_t);
                    }
                    break;
                default:
                    DEBUG_WMSGF("1. odd message", "1xn", dmsg.id);
            }
        }

        if (decoder.cmd & DECODER_CMD_PAUSE)
        {
            /* send confirm that we are paused */
            if (qres == OS_ERR_NONE)
            {
                pmsg.id = PLAYER_MSG_ID_DEC_PAUSED;
                PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_empty_t);
            }
        } else {
            /* send message that we resume playback */
            if (qres == OS_ERR_NONE)
            {
                pmsg.id = PLAYER_MSG_ID_DEC_PLAY;
                PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_empty_t);
            }
            break;
        }
    }

    return DECODER_PROCMSG_NOP;
}

/*
 * resume playback
 *
 * RETURN
 *     1 on success
 *     0 on error
 */
static int decoder_resume()
{
#define EFBUF_SIZE     32
    uint8 efbuf[EFBUF_SIZE];
    int volume;
    struct fcache_entry_t *entry;
    int drained;
    int ret;
    uint8 *buf;
    int n;

    ret = 1;

    volume = player.volume;

    /* set volume to minimum */
    vs1053b_set_low_fclk(); /* NOTE */
    vs1053b_set_volume(0);

    /*
     * 1. Send a portion of an audio file to VS1053b.
     * 2. When random access is required, read SCI_STATUS bit SS_DO_NOT_JUMP. If that bit
     *    is set, random access cannot be performed, so go back to 1.
     * 3. Read extra parameter value endFillByte (Chapter 10.11).
     * 4. Send at least 2048 bytes of endFillByte[7:0].
     * 5. Jump forwards or backwards in the file.
     * 6. Continue sending the file.    
     */

    /* precache */
    player_fcache_fill();

    vs1053b_set_hi_fclk(); /* NOTE */

    DEBUG_IMSG("send portion");
    while (1)
    {
        entry = player_fcache_get(&drained);
        if (!entry)
        {
            /*
             * If file was drained and cache underflow was occured we can stop
             * decoding.
             */
            if (drained)
            {
                ret = 0;
                goto out;
            }

            DEBUG_WMSG("cache underflow");

            /* fill cache */
            player_fcache_fill();
            continue;
        }

        buf = entry->buf;

        while (entry->len)
        {
            n = entry->len;
            if (n > FEED_PORTION)
                n = FEED_PORTION;

            /* 1. Send a portion of an audio file to VS1053b. */
            vs1053b_writesdi(buf, n);

            buf          += n;
            entry->len   -= n;
            decoder.fpos += n;

            if (!vs1053b_check_dreq())
            {
                vs1053b_set_low_fclk(); /* NOTE */

                /*
                 * 2. When random access is required, read SCI_STATUS bit SS_DO_NOT_JUMP. If that bit
                 *    is set, random access cannot be performed, so go back to 1.
                 */
                if (vs1053b_check_jump())
                    goto jump;
                vs1053b_set_hi_fclk(); /* NOTE */

                /* fill cache */
                player_fcache_fill();

                vs1053b_wait_dreq();
            }
        }
        player_fcache_remove();
    }
jump:
    /*
     * 3. Read extra parameter value endFillByte (Chapter 10.11).
     */

    DEBUG_IMSG("send endFillByte");

    /*
     * 4. Send at least 2048 bytes of endFillByte[7:0].
     */

    vs1053b_set_low_fclk(); /* NOTE */
    memset(efbuf, vs1053b_get_endfillbyte(), EFBUF_SIZE);
    vs1053b_set_hi_fclk(); /* NOTE */

    for (n = 0; n < (2048 / EFBUF_SIZE); n++)
    {
        vs1053b_writesdi(efbuf, EFBUF_SIZE);
        if (!vs1053b_check_dreq())
            vs1053b_wait_dreq();
    }

    /*
     * 5. Jump forwards or backwards in the file.
     */

    DEBUG_IMSGF("jump, position", "4dn", decoder.res_fpos);
    if (!player_fcache_seek(decoder.res_fpos))
    {
        ret = 0;
        goto out;
    }

    /*
     * 6. Continue sending the file.    
     */

    DEBUG_IMSG("continue");

    /* set decode time and file position */
    vs1053b_set_low_fclk(); /* NOTE */

    DEBUG_IMSGF("set time, time", "4dn", decoder.res_time);
    vs1053b_set_decode_time(decoder.res_time);
    decoder.fpos = decoder.res_fpos;

    decoder_send_position(DECODER_SEND_POSITION_FORCED);

    /*
     * XXX
     *     to switch play/pause icon - if paused send pause message to decoder,
     *     else send play message to player :)
     */

    if (decoder.res_pause)
    {
        struct decoder_msg_t dmsg;

        dmsg.id = DECODER_MSG_ID_PAUSE;
        DECODER_SEND_MSG(&dmsg, dmsg_empty_t);
    } else {
        struct player_msg_t pmsg;

        pmsg.id = PLAYER_MSG_ID_DEC_PLAY;
        PLAYER_SEND_MSG_BLOCKING(&pmsg, pmsg_empty_t);
    }
out:
    decoder_set_volume(volume);
    return ret;
}

