/* 
 *     This file is part of K11, hardware multimedia player.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * raster font support
 */
#include <debug.h>
#include <string.h>
#include "gs.h"

#define DEBUG_FONT

#ifdef DEBUG_FONT
    #define DPRINT(fmt, ...)  dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

#define FONT_7X8_SIZE     ( 7 *  8 * 256 / 8)
#define FONT_9X16_SIZE    ( 9 * 16 * 256 / 8)
#define FONT_8X16_SIZE    ( 8 * 16 * 256 / 8)
#define FONT_15X24_SIZE   (15 * 24 * (7 * 16) / 8)

extern const uint8 f7x8[FONT_7X8_SIZE];
//extern const uint8 f9x16[FONT_9X16_SIZE];
extern const uint8 f8x16[FONT_8X16_SIZE];
extern const uint8 f15x24[FONT_15X24_SIZE];

uint8 font7x8_data[FONT_7X8_SIZE];
//uint8 font9x16_data[FONT_9X16_SIZE];
uint8 font8x16_data[FONT_8X16_SIZE];
uint8 font15x24_data[FONT_15X24_SIZE];

/* raster fonts */
struct gs_font_t gs_fonts[GS_RASTER_FONT_COUNT] = {
    /* width     height number of symbols    offset   pointer to font data */
    {  7,         8,    256,                 0,       font7x8_data},
//    {  9,        16,    256,                 0,       font9x16_data},
    {  8,        16,    256,                 0,       font8x16_data},
    {  15,       24,    (7*16),              16,      font15x24_data},
};

/*
 * load data of fonts from external source
 */
void gs_font_init()
{
    /*
     * NOTE temporary use compiled font data
     */
    gs_font_set_font_data(GS_FONT_7X8,   (uint8 *)f7x8,   FONT_7X8_SIZE);
//    gs_font_set_font_data(GS_FONT_9X16,  (uint8 *)f9x16,  FONT_9X16_SIZE);
    gs_font_set_font_data(GS_FONT_8X16,  (uint8 *)f8x16,  FONT_8X16_SIZE);
    gs_font_set_font_data(GS_FONT_15X24, (uint8 *)f15x24, FONT_15X24_SIZE);
}

/*
 *
 */
void gs_font_set_font_data(struct gs_font_t *font, uint8 *data, int len)
{
    int dlen;

    dlen = font->width * font->height * font->nsym / 8;
    if (len != dlen)
    {
        DEBUG_EMSG("font data length mismatch");
        return;
    }
    memcpy(font->data, data, len);
}

/*
 * Binary font format
 * ===================

 * 7x8 character layout:

 *         byte1
 *           |    +--byte6
 *   byte0--+|    |
 *   |      ||    |
 *   |      0000000
 *   |      1111111
 *   |      2222222
 *   |      3333333
 *   |      4444444
 *   |      5555555
 *   |      6666666
 *   +------7777777

 *     
 * 9x16 character layout:

 *        byte2
 *           |
 *   byte0--+|
 *   |      ||
 *   |      000000000
 *   |      111111111
 *   |      222222222
 *   |      333333333
 *   |      444444444
 *   |      555555555
 *   |      666666666
 *   +------777777777
 *    +-----000000000
 *    |     111111111
 *    |     222222222
 *    |     333333333
 *    |     444444444
 *    |     555555555
 *    |     666666666
 *    |     777777777
 *    |     ||      |
 *   byte1--+|      +---byte17
 *   byte3---+
 */

const uint8 f7x8[FONT_7X8_SIZE] = {
    0x7F, 0x81, 0x81, 0x81, 0x81, 0x81, 0xFE, 0x00, 0x30, 0x3C, 0x7E, 0x3C, 0x30, 0x00, 0x00, 0x66, 
    0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 
    0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 
    0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 
    0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 
    0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 
    0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 
    0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 
    0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 
    0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 
    0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 
    0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 
    0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 
    0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5E, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x14, 0x7F, 0x14, 0x7F, 0x14, 0x00, 0x00, 0x04, 0x2A, 0x7F, 
    0x2A, 0x10, 0x00, 0x00, 0x00, 0x16, 0x08, 0x34, 0x00, 0x00, 0x00, 0x36, 0x49, 0x36, 0x40, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3C, 0x42, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x42, 0x3C, 0x00, 0x00, 0x00, 0x00, 0x54, 0x38, 0x38, 0x54, 0x00, 0x00, 0x00, 0x10, 0x10, 
    0x7C, 0x10, 0x10, 0x00, 0x00, 0x00, 0x80, 0x60, 0x20, 0x00, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 
    0x00, 0x00, 0x00, 0x00, 0x40, 0xE0, 0x40, 0x00, 0x00, 0x00, 0x60, 0x10, 0x08, 0x06, 0x00, 0x00, 
    0x00, 0x00, 0x3C, 0x42, 0x3C, 0x00, 0x00, 0x00, 0x00, 0x44, 0x7E, 0x40, 0x00, 0x00, 0x00, 0x64, 
    0x52, 0x52, 0x4C, 0x00, 0x00, 0x00, 0x22, 0x4A, 0x4E, 0x32, 0x00, 0x00, 0x00, 0x18, 0x14, 0x7E, 
    0x10, 0x00, 0x00, 0x00, 0x2E, 0x4A, 0x4A, 0x32, 0x00, 0x00, 0x00, 0x3C, 0x4A, 0x4A, 0x30, 0x00, 
    0x00, 0x00, 0x02, 0x62, 0x1A, 0x06, 0x00, 0x00, 0x00, 0x34, 0x4A, 0x4A, 0x34, 0x00, 0x00, 0x00, 
    0x0C, 0x52, 0x52, 0x3C, 0x00, 0x00, 0x00, 0x00, 0x6C, 0x6C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 
    0x6C, 0x2C, 0x00, 0x00, 0x00, 0x00, 0x18, 0x24, 0x42, 0x00, 0x00, 0x00, 0x28, 0x28, 0x28, 0x28, 
    0x00, 0x00, 0x00, 0x00, 0x42, 0x24, 0x18, 0x00, 0x00, 0x00, 0x00, 0x04, 0x52, 0x0C, 0x00, 0x00, 
    0x00, 0x3C, 0x42, 0x99, 0xA5, 0x1E, 0x00, 0x00, 0x7C, 0x12, 0x12, 0x7C, 0x00, 0x00, 0x00, 0x7E, 
    0x4A, 0x4A, 0x34, 0x00, 0x00, 0x00, 0x3C, 0x42, 0x42, 0x24, 0x00, 0x00, 0x00, 0x7E, 0x42, 0x42, 
    0x3C, 0x00, 0x00, 0x00, 0x7E, 0x4A, 0x4A, 0x42, 0x00, 0x00, 0x00, 0x7E, 0x0A, 0x0A, 0x02, 0x00, 
    0x00, 0x00, 0x3C, 0x42, 0x52, 0x34, 0x00, 0x00, 0x00, 0x7E, 0x08, 0x08, 0x7E, 0x00, 0x00, 0x00, 
    0x00, 0x42, 0x7E, 0x42, 0x00, 0x00, 0x00, 0x20, 0x42, 0x3E, 0x02, 0x00, 0x00, 0x00, 0x7E, 0x08, 
    0x34, 0x42, 0x00, 0x00, 0x00, 0x7E, 0x40, 0x40, 0x40, 0x00, 0x00, 0x00, 0x7E, 0x0C, 0x0C, 0x7E, 
    0x00, 0x00, 0x00, 0x7E, 0x0C, 0x38, 0x7E, 0x00, 0x00, 0x00, 0x3C, 0x42, 0x42, 0x3C, 0x00, 0x00, 
    0x00, 0x7E, 0x12, 0x12, 0x0C, 0x00, 0x00, 0x00, 0x3C, 0x52, 0x62, 0xBC, 0x00, 0x00, 0x00, 0x7E, 
    0x12, 0x12, 0x6C, 0x00, 0x00, 0x00, 0x24, 0x4A, 0x52, 0x24, 0x00, 0x00, 0x00, 0x00, 0x02, 0x7E, 
    0x02, 0x00, 0x00, 0x00, 0x3E, 0x40, 0x40, 0x3E, 0x00, 0x00, 0x00, 0x1E, 0x60, 0x60, 0x1E, 0x00, 
    0x00, 0x00, 0x7E, 0x30, 0x30, 0x7E, 0x00, 0x00, 0x00, 0x66, 0x18, 0x18, 0x66, 0x00, 0x00, 0x00, 
    0x06, 0x08, 0x70, 0x08, 0x06, 0x00, 0x00, 0x62, 0x52, 0x4A, 0x46, 0x00, 0x00, 0x00, 0x00, 0x7E, 
    0x42, 0x42, 0x00, 0x00, 0x00, 0x06, 0x08, 0x10, 0x60, 0x00, 0x00, 0x00, 0x00, 0x42, 0x42, 0x7E, 
    0x00, 0x00, 0x00, 0x00, 0x04, 0x02, 0x04, 0x00, 0x00, 0x00, 0x80, 0x80, 0x80, 0x80, 0x00, 0x00, 
    0x00, 0x00, 0x02, 0x04, 0x00, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x78, 0x00, 0x00, 0x00, 0x7E, 
    0x48, 0x48, 0x30, 0x00, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 
    0x7E, 0x00, 0x00, 0x00, 0x30, 0x68, 0x58, 0x10, 0x00, 0x00, 0x00, 0x10, 0x7C, 0x12, 0x04, 0x00, 
    0x00, 0x00, 0x10, 0xA8, 0xA8, 0x70, 0x00, 0x00, 0x00, 0x7E, 0x08, 0x08, 0x70, 0x00, 0x00, 0x00, 
    0x00, 0x48, 0x7A, 0x40, 0x00, 0x00, 0x00, 0x00, 0x40, 0x80, 0x7A, 0x00, 0x00, 0x00, 0x7E, 0x10, 
    0x10, 0x68, 0x00, 0x00, 0x00, 0x00, 0x42, 0x7E, 0x40, 0x00, 0x00, 0x00, 0x78, 0x08, 0x70, 0x08, 
    0x70, 0x00, 0x00, 0x78, 0x08, 0x08, 0x70, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x30, 0x00, 0x00, 
    0x00, 0xF8, 0x28, 0x28, 0x10, 0x00, 0x00, 0x00, 0x10, 0x28, 0x28, 0xF8, 0x00, 0x00, 0x00, 0x78, 
    0x10, 0x08, 0x10, 0x00, 0x00, 0x00, 0x00, 0x50, 0x58, 0x28, 0x00, 0x00, 0x00, 0x08, 0x3E, 0x48, 
    0x20, 0x00, 0x00, 0x00, 0x38, 0x40, 0x40, 0x78, 0x00, 0x00, 0x00, 0x00, 0x38, 0x40, 0x38, 0x00, 
    0x00, 0x00, 0x38, 0x40, 0x30, 0x40, 0x38, 0x00, 0x00, 0x48, 0x30, 0x30, 0x48, 0x00, 0x00, 0x00, 
    0x58, 0xA0, 0xA0, 0x78, 0x00, 0x00, 0x00, 0x48, 0x68, 0x58, 0x48, 0x00, 0x00, 0x00, 0x08, 0x2A, 
    0x55, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x41, 0x55, 0x2A, 0x08, 
    0x00, 0x00, 0x00, 0x04, 0x02, 0x04, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0xF8, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0xF8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0F, 
    0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x08, 0x08, 
    0x08, 0x08, 0x08, 0x08, 0xFF, 0x00, 0x00, 0x00, 0x08, 0x08, 0x08, 0xF8, 0x08, 0x08, 0x08, 0x08, 
    0x08, 0x08, 0x0F, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0xFF, 0x08, 0x08, 0x08, 0x0F, 0x0F, 0x0F, 
    0x0F, 0x0F, 0x0F, 0x0F, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
    0x44, 0x11, 0x44, 0x11, 0x44, 0x11, 0x44, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x00, 0xBB, 0xEE, 
    0xBB, 0xEE, 0xBB, 0xEE, 0xFF, 0x00, 0x00, 0x00, 0xFC, 0x02, 0x04, 0x00, 0x00, 0x78, 0x78, 0x78, 
    0x78, 0x00, 0x00, 0x00, 0x00, 0x38, 0x38, 0x38, 0x00, 0x00, 0x00, 0x10, 0x30, 0xC0, 0x3F, 0x01, 
    0x00, 0x00, 0x50, 0x28, 0x50, 0x28, 0x00, 0x00, 0x00, 0x00, 0x48, 0x54, 0x62, 0x00, 0x00, 0x00, 
    0x00, 0x62, 0x54, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x40, 
    0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x0A, 0x04, 0x00, 0x00, 0x00, 0x00, 0x12, 0x19, 0x16, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x54, 0x10, 0x00, 0x00, 
    0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0xFC, 0x14, 0x14, 0x14, 0x00, 0x32, 0x68, 0x5A, 0x10, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x08, 
    0xF8, 0x08, 0x08, 0x00, 0x00, 0xFC, 0x04, 0xF4, 0x14, 0x14, 0x14, 0x14, 0x14, 0xFC, 0x00, 0x00, 
    0x00, 0x08, 0x08, 0xF8, 0x08, 0xF8, 0x00, 0x00, 0x14, 0x14, 0xF4, 0x04, 0xFC, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x1F, 0x14, 0x14, 0x14, 0x00, 0x00, 0x0F, 0x08, 0x0F, 0x08, 0x08, 0x00, 0x00, 0x1F, 
    0x10, 0x17, 0x14, 0x14, 0x14, 0x14, 0x14, 0x1F, 0x00, 0x00, 0x00, 0x08, 0x08, 0x0F, 0x08, 0x0F, 
    0x00, 0x00, 0x14, 0x14, 0x17, 0x10, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x14, 0x14, 0x14, 
    0x00, 0x00, 0xFF, 0x00, 0xFF, 0x08, 0x08, 0x00, 0x00, 0xFF, 0x00, 0xF7, 0x14, 0x14, 0x14, 0x14, 
    0x14, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x4A, 0x4A, 0x42, 0x00, 0x00, 0x08, 0x08, 0xFF, 0x00, 
    0xFF, 0x00, 0x00, 0x14, 0x14, 0xF7, 0x00, 0xFF, 0x00, 0x00, 0x14, 0x14, 0x14, 0xF4, 0x14, 0x14, 
    0x14, 0x08, 0x08, 0xF8, 0x08, 0xF8, 0x08, 0x08, 0x14, 0x14, 0xF4, 0x04, 0xF4, 0x14, 0x14, 0x14, 
    0x14, 0x14, 0x17, 0x14, 0x14, 0x14, 0x08, 0x08, 0x0F, 0x08, 0x0F, 0x08, 0x08, 0x14, 0x14, 0x17, 
    0x10, 0x17, 0x14, 0x14, 0x14, 0x14, 0x14, 0xFF, 0x14, 0x14, 0x14, 0x08, 0x08, 0xFF, 0x08, 0xFF, 
    0x08, 0x08, 0x14, 0x14, 0xF7, 0x00, 0xF7, 0x14, 0x14, 0x00, 0x3C, 0x5A, 0x66, 0x42, 0x3C, 0x00, 
    0x00, 0x78, 0x30, 0x48, 0x30, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x78, 0x00, 0x00, 0x00, 0x3C, 
    0x4A, 0x4A, 0x30, 0x00, 0x00, 0x00, 0x78, 0x40, 0x40, 0xF8, 0x00, 0x00, 0x00, 0xE0, 0x58, 0x48, 
    0xF8, 0x00, 0x00, 0x00, 0x30, 0x68, 0x58, 0x10, 0x00, 0x00, 0x00, 0x20, 0x50, 0xF8, 0x50, 0x20, 
    0x00, 0x00, 0x00, 0x78, 0x08, 0x08, 0x00, 0x00, 0x00, 0x48, 0x30, 0x30, 0x48, 0x00, 0x00, 0x00, 
    0x78, 0x20, 0x10, 0x78, 0x00, 0x00, 0x00, 0x7A, 0x24, 0x14, 0x7A, 0x00, 0x00, 0x00, 0x78, 0x10, 
    0x20, 0x48, 0x00, 0x00, 0x00, 0x40, 0x30, 0x08, 0x78, 0x00, 0x00, 0x00, 0x78, 0x10, 0x20, 0x10, 
    0x78, 0x00, 0x00, 0x78, 0x10, 0x10, 0x78, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x30, 0x00, 0x00, 
    0x00, 0x78, 0x08, 0x08, 0x78, 0x00, 0x00, 0x00, 0x00, 0x50, 0x28, 0x78, 0x00, 0x00, 0x00, 0xF8, 
    0x48, 0x48, 0x30, 0x00, 0x00, 0x00, 0x00, 0x30, 0x48, 0x48, 0x00, 0x00, 0x00, 0x00, 0x08, 0x78, 
    0x08, 0x00, 0x00, 0x00, 0x18, 0xA0, 0x40, 0x38, 0x00, 0x00, 0x00, 0x68, 0x10, 0x78, 0x10, 0x68, 
    0x00, 0x00, 0x7C, 0x54, 0x58, 0x20, 0x00, 0x00, 0x00, 0x78, 0x50, 0x50, 0x20, 0x00, 0x00, 0x00, 
    0x78, 0x50, 0x20, 0x78, 0x00, 0x00, 0x00, 0x00, 0x48, 0x58, 0x38, 0x00, 0x00, 0x00, 0x78, 0x40, 
    0x78, 0x40, 0x78, 0x00, 0x00, 0x00, 0x48, 0x58, 0x30, 0x00, 0x00, 0x00, 0x78, 0x40, 0x78, 0x40, 
    0xF8, 0x00, 0x00, 0x18, 0x20, 0x20, 0x78, 0x00, 0x00, 0x00, 0x08, 0x78, 0x50, 0x20, 0x00, 0x00, 
    0x00, 0x7E, 0x08, 0x3C, 0x42, 0x3C, 0x00, 0x00, 0x7C, 0x12, 0x12, 0x7C, 0x00, 0x00, 0x00, 0x7E, 
    0x4A, 0x4A, 0x30, 0x00, 0x00, 0x00, 0x7E, 0x40, 0x7E, 0xC0, 0x00, 0x00, 0x00, 0xFC, 0x42, 0x7E, 
    0xC0, 0x00, 0x00, 0x00, 0x7E, 0x4A, 0x4A, 0x42, 0x00, 0x00, 0x00, 0x18, 0x24, 0x7E, 0x24, 0x18, 
    0x00, 0x00, 0x7E, 0x02, 0x02, 0x06, 0x00, 0x00, 0x00, 0x00, 0x66, 0x18, 0x66, 0x00, 0x00, 0x00, 
    0x7E, 0x10, 0x08, 0x7E, 0x00, 0x00, 0x00, 0x7D, 0x12, 0x0A, 0x7D, 0x00, 0x00, 0x00, 0x7E, 0x08, 
    0x14, 0x62, 0x00, 0x00, 0x00, 0x40, 0x3C, 0x02, 0x7E, 0x00, 0x00, 0x00, 0x7E, 0x0C, 0x0C, 0x7E, 
    0x00, 0x00, 0x00, 0x7E, 0x08, 0x08, 0x7E, 0x00, 0x00, 0x00, 0x3C, 0x42, 0x42, 0x3C, 0x00, 0x00, 
    0x00, 0x7E, 0x02, 0x02, 0x7E, 0x00, 0x00, 0x00, 0x6C, 0x12, 0x12, 0x7E, 0x00, 0x00, 0x00, 0x7E, 
    0x12, 0x12, 0x0C, 0x00, 0x00, 0x00, 0x3C, 0x42, 0x42, 0x24, 0x00, 0x00, 0x00, 0x00, 0x02, 0x7E, 
    0x02, 0x00, 0x00, 0x00, 0x4E, 0x50, 0x50, 0x3E, 0x00, 0x00, 0x00, 0x76, 0x08, 0x7E, 0x08, 0x76, 
    0x00, 0x00, 0x7E, 0x4A, 0x4A, 0x34, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x48, 0x30, 0x00, 0x00, 0x00, 
    0x7E, 0x48, 0x30, 0x00, 0x7E, 0x00, 0x00, 0x24, 0x42, 0x4A, 0x34, 0x00, 0x00, 0x00, 0x7E, 0x40, 
    0x7E, 0x40, 0x7E, 0x00, 0x00, 0x42, 0x4A, 0x4A, 0x3C, 0x00, 0x00, 0x00, 0x7E, 0x40, 0x7E, 0x40, 
    0xFE, 0x00, 0x00, 0x00, 0x0E, 0x10, 0x7E, 0x00, 0x00, 0x00, 0x02, 0x02, 0x7E, 0x48, 0x30, 0x00, 
};

const uint8 f8x16[] = {
    /*0000*/    0xFF, 0x7F, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFE, 0x7F, 
    /*0010*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0020*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0030*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0040*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0050*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0060*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0070*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0080*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0090*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00A0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00B0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00C0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00D0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00E0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*00F0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0100*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0110*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0120*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0130*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0140*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0150*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0160*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0170*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0180*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0190*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01A0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01B0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01C0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01D0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01E0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*01F0*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0200*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0210*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0D, 0xFC, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0220*/    0x00, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 
    /*0230*/    0x20, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x01, 0x00, 0x00, 
    /*0240*/    0x70, 0x04, 0xF8, 0x0C, 0x88, 0x08, 0xFE, 0x3F, 0x88, 0x08, 0x98, 0x0F, 0x10, 0x07, 0x00, 0x00, 
    /*0250*/    0x08, 0x00, 0x1C, 0x0C, 0x14, 0x0F, 0xC8, 0x03, 0xF0, 0x04, 0x3C, 0x0A, 0x0C, 0x0E, 0x00, 0x04, 
    /*0260*/    0x80, 0x07, 0xD8, 0x0F, 0x7C, 0x08, 0xE4, 0x0C, 0xBC, 0x07, 0xD8, 0x0F, 0x40, 0x08, 0x00, 0x00, 
    /*0270*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0280*/    0x00, 0x00, 0x00, 0x00, 0xF0, 0x03, 0xF8, 0x07, 0x0C, 0x0C, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*0290*/    0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0x0C, 0x0C, 0xF8, 0x07, 0xF0, 0x03, 0x00, 0x00, 0x00, 0x00, 
    /*02A0*/    0x80, 0x00, 0xA0, 0x02, 0xE0, 0x03, 0xC0, 0x01, 0xE0, 0x03, 0xA0, 0x02, 0x80, 0x00, 0x00, 0x00, 
    /*02B0*/    0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0xE0, 0x03, 0xE0, 0x03, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 
    /*02C0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x1C, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*02D0*/    0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 
    /*02E0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*02F0*/    0x00, 0x00, 0x00, 0x0C, 0x00, 0x0F, 0xC0, 0x03, 0xF0, 0x00, 0x3C, 0x00, 0x0C, 0x00, 0x00, 0x00, 
    /*0300*/    0xF8, 0x07, 0xFC, 0x0F, 0x84, 0x09, 0xC4, 0x08, 0x64, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*0310*/    0x00, 0x00, 0x10, 0x08, 0x18, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x00, 
    /*0320*/    0x18, 0x0C, 0x1C, 0x0E, 0x04, 0x0B, 0x84, 0x09, 0xC4, 0x08, 0x7C, 0x08, 0x38, 0x08, 0x00, 0x00, 
    /*0330*/    0x18, 0x06, 0x1C, 0x0E, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
    /*0340*/    0x80, 0x01, 0xC0, 0x01, 0x60, 0x01, 0x30, 0x01, 0x18, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0350*/    0x7C, 0x04, 0x7C, 0x0C, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xC4, 0x0F, 0x84, 0x07, 0x00, 0x00, 
    /*0360*/    0xF0, 0x07, 0xF8, 0x0F, 0x4C, 0x08, 0x44, 0x08, 0x44, 0x08, 0xC4, 0x0F, 0x80, 0x07, 0x00, 0x00, 
    /*0370*/    0x04, 0x00, 0x04, 0x00, 0x04, 0x0E, 0x84, 0x0F, 0xE4, 0x01, 0x7C, 0x00, 0x1C, 0x00, 0x00, 0x00, 
    /*0380*/    0xB8, 0x07, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
    /*0390*/    0x78, 0x00, 0xFC, 0x08, 0x84, 0x08, 0x84, 0x08, 0x84, 0x0C, 0xFC, 0x07, 0xF8, 0x03, 0x00, 0x00, 
    /*03A0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x0C, 0x60, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*03B0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x60, 0x1C, 0x60, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*03C0*/    0x00, 0x00, 0x80, 0x00, 0xC0, 0x01, 0x60, 0x03, 0x30, 0x06, 0x18, 0x0C, 0x08, 0x08, 0x00, 0x00, 
    /*03D0*/    0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x00, 0x00, 
    /*03E0*/    0x00, 0x00, 0x08, 0x08, 0x18, 0x0C, 0x30, 0x06, 0x60, 0x03, 0xC0, 0x01, 0x80, 0x00, 0x00, 0x00, 
    /*03F0*/    0x38, 0x00, 0x3C, 0x00, 0x04, 0x00, 0x84, 0x0D, 0xC4, 0x0D, 0x7C, 0x00, 0x38, 0x00, 0x00, 0x00, 
    /*0400*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0xE4, 0x09, 0x14, 0x0A, 0xFC, 0x0B, 0xF8, 0x0B, 0x00, 0x00, 
    /*0410*/    0xF8, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x0F, 0xF8, 0x0F, 0x00, 0x00, 
    /*0420*/    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
    /*0430*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x1C, 0x0E, 0x18, 0x06, 0x00, 0x00, 
    /*0440*/    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x0C, 0x0C, 0xF8, 0x07, 0xF0, 0x03, 0x00, 0x00, 
    /*0450*/    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 
    /*0460*/    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x00, 0x44, 0x00, 0x44, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 
    /*0470*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x84, 0x08, 0x84, 0x08, 0x9C, 0x0F, 0x98, 0x07, 0x00, 0x00, 
    /*0480*/    0xFC, 0x0F, 0xFC, 0x0F, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0490*/    0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*04A0*/    0x00, 0x06, 0x00, 0x0E, 0x00, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x07, 0x04, 0x00, 0x00, 0x00, 
    /*04B0*/    0xFC, 0x0F, 0xFC, 0x0F, 0xC0, 0x00, 0xE0, 0x01, 0x30, 0x03, 0x1C, 0x0E, 0x0C, 0x0C, 0x00, 0x00, 
    /*04C0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00, 0x00, 
    /*04D0*/    0xFC, 0x0F, 0xF8, 0x0F, 0x30, 0x00, 0x60, 0x00, 0x30, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*04E0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x60, 0x00, 0xC0, 0x00, 0x80, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*04F0*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*0500*/    0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x00, 0x78, 0x00, 0x00, 0x00, 
    /*0510*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x0C, 0x04, 0x0C, 0xFC, 0x1F, 0xF8, 0x17, 0x00, 0x00, 
    /*0520*/    0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x01, 0x84, 0x03, 0x84, 0x06, 0xFC, 0x0C, 0x78, 0x08, 0x00, 0x00, 
    /*0530*/    0x38, 0x06, 0x7C, 0x0E, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xCC, 0x0F, 0x88, 0x07, 0x00, 0x00, 
    /*0540*/    0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 
    /*0550*/    0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x00, 
    /*0560*/    0x7C, 0x00, 0xFC, 0x03, 0x80, 0x0F, 0x00, 0x0C, 0x80, 0x0F, 0xFC, 0x03, 0x7C, 0x00, 0x00, 0x00, 
    /*0570*/    0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x03, 0x80, 0x01, 0x00, 0x03, 0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x00, 
    /*0580*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0590*/    0x0C, 0x00, 0x3C, 0x00, 0x70, 0x00, 0xC0, 0x0F, 0xC0, 0x0F, 0x70, 0x00, 0x3C, 0x00, 0x0C, 0x00, 
    /*05A0*/    0x04, 0x0E, 0x04, 0x0F, 0x84, 0x09, 0xC4, 0x08, 0x64, 0x08, 0x3C, 0x08, 0x1C, 0x08, 0x00, 0x00, 
    /*05B0*/    0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*05C0*/    0x00, 0x00, 0x0C, 0x00, 0x3C, 0x00, 0xF0, 0x00, 0xC0, 0x03, 0x00, 0x0F, 0x00, 0x0C, 0x00, 0x00, 
    /*05D0*/    0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 
    /*05E0*/    0x00, 0x00, 0x08, 0x00, 0x0C, 0x00, 0x06, 0x00, 0x06, 0x00, 0x0C, 0x00, 0x08, 0x00, 0x00, 0x00, 
    /*05F0*/    0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x00, 
    /*0600*/    0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x03, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0610*/    0x00, 0x07, 0xA0, 0x0F, 0xA0, 0x08, 0xA0, 0x08, 0xA0, 0x08, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 
    /*0620*/    0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0630*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0x60, 0x0C, 0x40, 0x04, 0x00, 0x00, 
    /*0640*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0650*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x01, 0x00, 0x00, 
    /*0660*/    0x20, 0x00, 0x20, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x24, 0x00, 0x24, 0x00, 0x04, 0x00, 0x00, 0x00, 
    /*0670*/    0xC0, 0x07, 0xE0, 0x4F, 0x20, 0x48, 0x20, 0x48, 0x20, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
    /*0680*/    0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 
    /*0690*/    0x00, 0x00, 0x00, 0x00, 0x20, 0x08, 0xEC, 0x0F, 0xEC, 0x0F, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*06A0*/    0x00, 0x00, 0x00, 0x30, 0x00, 0x70, 0x00, 0x40, 0x20, 0x40, 0xEC, 0x7F, 0xEC, 0x3F, 0x00, 0x00, 
    /*06B0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x01, 0x80, 0x03, 0xC0, 0x06, 0x60, 0x0C, 0x20, 0x08, 0x00, 0x00, 
    /*06C0*/    0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*06D0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0xE0, 0x0F, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 
    /*06E0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 
    /*06F0*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0700*/    0xE0, 0x7F, 0xE0, 0x7F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0710*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x7F, 0xE0, 0x7F, 0x00, 0x00, 
    /*0720*/    0xE0, 0x0F, 0xE0, 0x0F, 0xC0, 0x00, 0x60, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 
    /*0730*/    0xC0, 0x08, 0xE0, 0x09, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0x20, 0x0F, 0x20, 0x06, 0x00, 0x00, 
    /*0740*/    0x20, 0x00, 0x20, 0x00, 0xFC, 0x07, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x00, 0x08, 0x00, 0x00, 
    /*0750*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0760*/    0xE0, 0x00, 0xE0, 0x03, 0x00, 0x0F, 0x00, 0x0C, 0x00, 0x0F, 0xE0, 0x03, 0xE0, 0x00, 0x00, 0x00, 
    /*0770*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x80, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x07, 0x00, 0x00, 
    /*0780*/    0x60, 0x0C, 0xE0, 0x0E, 0x80, 0x03, 0x00, 0x01, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 0x00, 0x00, 
    /*0790*/    0xE0, 0x07, 0xE0, 0x4F, 0x00, 0x48, 0x00, 0x48, 0x00, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
    /*07A0*/    0x20, 0x0C, 0x20, 0x0E, 0x20, 0x0B, 0xA0, 0x09, 0xE0, 0x08, 0x60, 0x08, 0x20, 0x08, 0x00, 0x00, 
    /*07B0*/    0x00, 0x00, 0x40, 0x00, 0xF8, 0x07, 0xBC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*07C0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07D0*/    0x00, 0x00, 0x04, 0x08, 0x04, 0x08, 0xBC, 0x0F, 0xF8, 0x07, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07E0*/    0x0C, 0x00, 0x0E, 0x00, 0x02, 0x00, 0x06, 0x00, 0x0C, 0x00, 0x08, 0x00, 0x0E, 0x00, 0x06, 0x00, 
    /*07F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0800*/    0xF8, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x0F, 0xF8, 0x0F, 0x00, 0x00, 
    /*0810*/    0xFC, 0x0F, 0xFC, 0x0F, 0x24, 0x08, 0x24, 0x08, 0x24, 0x08, 0xE4, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0820*/    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
    /*0830*/    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 
    /*0840*/    0x00, 0x18, 0xF8, 0x1F, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x1F, 0x00, 0x18, 
    /*0850*/    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 
    /*0860*/    0x1C, 0x0F, 0xBC, 0x0F, 0xE0, 0x00, 0xFC, 0x0F, 0xE0, 0x00, 0xBC, 0x0F, 0x1C, 0x0F, 0x00, 0x00, 
    /*0870*/    0x18, 0x06, 0x1C, 0x0E, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
    /*0880*/    0xFC, 0x0F, 0xFC, 0x0F, 0x80, 0x01, 0xC0, 0x00, 0x60, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0890*/    0xFC, 0x0F, 0xFD, 0x0F, 0x83, 0x01, 0xC2, 0x00, 0x63, 0x00, 0xFD, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*08A0*/    0xFC, 0x0F, 0xFC, 0x0F, 0xC0, 0x00, 0xE0, 0x01, 0x30, 0x03, 0x1C, 0x0E, 0x0C, 0x0C, 0x00, 0x00, 
    /*08B0*/    0x00, 0x08, 0xF0, 0x0F, 0xF8, 0x07, 0x0C, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*08C0*/    0xFC, 0x0F, 0xF8, 0x0F, 0x30, 0x00, 0x60, 0x00, 0x30, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*08D0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*08E0*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*08F0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0900*/    0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x00, 0x78, 0x00, 0x00, 0x00, 
    /*0910*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x1C, 0x0E, 0x18, 0x06, 0x00, 0x00, 
    /*0920*/    0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 
    /*0930*/    0x7C, 0x00, 0xFC, 0x08, 0x80, 0x08, 0x80, 0x08, 0x80, 0x08, 0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x00, 
    /*0940*/    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0xFE, 0x1F, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*0950*/    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
    /*0960*/    0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x3F, 0x00, 0x38, 
    /*0970*/    0x7C, 0x00, 0xFC, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0980*/    0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0990*/    0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x3F, 0x00, 0x38, 
    /*09A0*/    0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*09B0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0xC0, 0x07, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*09C0*/    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*09D0*/    0x08, 0x04, 0x0C, 0x0C, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*09E0*/    0xFC, 0x0F, 0xFC, 0x0F, 0x80, 0x00, 0xF8, 0x07, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
    /*09F0*/    0x78, 0x08, 0xFC, 0x0C, 0x84, 0x06, 0x84, 0x03, 0x84, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 
    /*0A00*/    0x00, 0x07, 0xA0, 0x0F, 0xA0, 0x08, 0xA0, 0x08, 0xA0, 0x08, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 
    /*0A10*/    0xF8, 0x0F, 0xFC, 0x0F, 0x24, 0x08, 0x24, 0x08, 0x24, 0x08, 0xE4, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0A20*/    0xF8, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x7C, 0x08, 0xD8, 0x0F, 0x80, 0x07, 0x00, 0x00, 
    /*0A30*/    0x00, 0x06, 0x20, 0x0F, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x00, 0x00, 0x00, 
    /*0A40*/    0xC0, 0x07, 0xE0, 0x4F, 0x20, 0x48, 0x20, 0x48, 0x20, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
    /*0A50*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x01, 0x00, 0x00, 
    /*0A60*/    0x60, 0x0C, 0xE0, 0x0E, 0x80, 0x03, 0xE0, 0x0F, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 0x00, 0x00, 
    /*0A70*/    0x40, 0x04, 0x60, 0x0C, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x0F, 0xC0, 0x06, 0x00, 0x00, 
    /*0A80*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0A90*/    0xE0, 0x07, 0xE4, 0x0F, 0x0C, 0x08, 0x08, 0x08, 0x0C, 0x08, 0xE4, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0AA0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x01, 0x80, 0x03, 0xC0, 0x06, 0x60, 0x0C, 0x20, 0x08, 0x00, 0x00, 
    /*0AB0*/    0x00, 0x08, 0xC0, 0x0F, 0xE0, 0x07, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0AC0*/    0xE0, 0x0F, 0xE0, 0x0F, 0xC0, 0x00, 0x80, 0x01, 0xC0, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0AD0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0AE0*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0AF0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0B00*/    0x55, 0x55, 0x00, 0x00, 0xAA, 0xAA, 0x00, 0x00, 0x55, 0x55, 0x00, 0x00, 0xAA, 0xAA, 0x00, 0x00, 
    /*0B10*/    0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 
    /*0B20*/    0xFF, 0xFF, 0x55, 0x55, 0xFF, 0xFF, 0xAA, 0xAA, 0xFF, 0xFF, 0x55, 0x55, 0xFF, 0xFF, 0xAA, 0xAA, 
    /*0B30*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B40*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B50*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B60*/    0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*0B70*/    0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*0B80*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xE0, 0xFF, 0xE0, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B90*/    0x60, 0x03, 0x7F, 0xFF, 0x7F, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*0BA0*/    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*0BB0*/    0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x00, 0xE0, 0xFF, 0xE0, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*0BC0*/    0x60, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x00, 0x03, 0xFF, 0x03, 0xFF, 0x03, 0x00, 0x00, 0x00, 0x00, 
    /*0BD0*/    0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 
    /*0BE0*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xFF, 0x03, 0xFF, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0BF0*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0C00*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C10*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C20*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C30*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C40*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C50*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0C60*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0C70*/    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 
    /*0C80*/    0x00, 0x00, 0xFF, 0x03, 0xFF, 0x03, 0x00, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0C90*/    0x00, 0x00, 0xE0, 0xFF, 0xE0, 0xFF, 0x60, 0x00, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 0x60, 0x03, 
    /*0CA0*/    0x60, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x00, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0CB0*/    0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x00, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 0x60, 0x03, 
    /*0CC0*/    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x7F, 0xFF, 0x7F, 0xFF, 0x60, 0x03, 0x60, 0x03, 
    /*0CD0*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0CE0*/    0x60, 0x03, 0x7F, 0xFF, 0x7F, 0xFF, 0x00, 0x00, 0x7F, 0xFF, 0x7F, 0xFF, 0x60, 0x03, 0x60, 0x03, 
    /*0CF0*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0D00*/    0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0D10*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0D20*/    0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 
    /*0D30*/    0x00, 0x00, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0D40*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x03, 0xFF, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0D50*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0xE0, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0D60*/    0x00, 0x00, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 
    /*0D70*/    0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 
    /*0D80*/    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
    /*0D90*/    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0DA0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
    /*0DB0*/    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
    /*0DC0*/    0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 
    /*0DD0*/    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0DE0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
    /*0DF0*/    0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 
    /*0E00*/    0xE0, 0x7F, 0xE0, 0x7F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0E10*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0x60, 0x0C, 0x40, 0x04, 0x00, 0x00, 
    /*0E20*/    0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 
    /*0E30*/    0xE0, 0x07, 0xE0, 0x4F, 0x00, 0x48, 0x00, 0x48, 0x00, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
    /*0E40*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0xF0, 0x1F, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0E50*/    0x60, 0x0C, 0xE0, 0x0E, 0x80, 0x03, 0x00, 0x01, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 0x00, 0x00, 
    /*0E60*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x3F, 0x00, 0x38, 
    /*0E70*/    0xE0, 0x00, 0xE0, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0E80*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0E90*/    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x3F, 0x00, 0x38, 
    /*0EA0*/    0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x80, 0x08, 0x80, 0x0F, 0x00, 0x07, 0x00, 0x00, 
    /*0EB0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x00, 0x07, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0EC0*/    0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x80, 0x08, 0x80, 0x0F, 0x00, 0x07, 0x00, 0x00, 
    /*0ED0*/    0x40, 0x04, 0x60, 0x0C, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0EE0*/    0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x01, 0xC0, 0x07, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
    /*0EF0*/    0xC0, 0x08, 0xE0, 0x0D, 0x20, 0x07, 0x20, 0x03, 0x20, 0x01, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 
    /*0F00*/    0xF8, 0x0F, 0xFB, 0x0F, 0x8B, 0x08, 0x88, 0x08, 0x8B, 0x08, 0x0B, 0x08, 0x08, 0x08, 0x00, 0x00, 
    /*0F10*/    0xC0, 0x07, 0xEC, 0x0F, 0x2C, 0x09, 0x20, 0x09, 0x2C, 0x09, 0xEC, 0x09, 0xC0, 0x01, 0x00, 0x00, 
    /*0F20*/    0xF8, 0x07, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0x0C, 0x0C, 0x08, 0x04, 0x00, 0x00, 
    /*0F30*/    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0x60, 0x0C, 0x40, 0x04, 0x00, 0x00, 
    /*0F40*/    0x00, 0x00, 0x03, 0x00, 0x0B, 0x08, 0xF8, 0x0F, 0xF8, 0x0F, 0x0B, 0x08, 0x03, 0x00, 0x00, 0x00, 
    /*0F50*/    0x00, 0x00, 0x0C, 0x00, 0x2C, 0x08, 0xE0, 0x0F, 0xEC, 0x0F, 0x0C, 0x08, 0x00, 0x00, 0x00, 0x00, 
    /*0F60*/    0x7C, 0x00, 0xFD, 0x08, 0x83, 0x08, 0x82, 0x08, 0x83, 0x08, 0xFD, 0x0F, 0xFC, 0x07, 0x00, 0x00, 
    /*0F70*/    0xE0, 0x07, 0xE4, 0x4F, 0x0C, 0x48, 0x08, 0x48, 0x0C, 0x48, 0xE4, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
    /*0F80*/    0x00, 0x00, 0x0C, 0x00, 0x1E, 0x00, 0x12, 0x00, 0x1E, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0F90*/    0x00, 0x00, 0x00, 0x00, 0xC0, 0x01, 0xC0, 0x01, 0xC0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0FA0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0FB0*/    0xC0, 0x01, 0xC0, 0x03, 0x00, 0x06, 0x00, 0x0C, 0xFE, 0x0F, 0xFE, 0x0F, 0x02, 0x00, 0x00, 0x00, 
    /*0FC0*/    0xFC, 0x0F, 0xE0, 0x00, 0xC0, 0x01, 0xFC, 0x0F, 0x00, 0x00, 0x1C, 0x0A, 0x1C, 0x0A, 0x00, 0x00, 
    /*0FD0*/    0x00, 0x00, 0xD0, 0x05, 0xF0, 0x07, 0x20, 0x02, 0x20, 0x02, 0xF0, 0x07, 0xD0, 0x05, 0x00, 0x00, 
    /*0FE0*/    0x00, 0x00, 0xE0, 0x07, 0xE0, 0x07, 0xE0, 0x07, 0xE0, 0x07, 0xE0, 0x07, 0x00, 0x00, 0x00, 0x00, 
    /*0FF0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
};


//const uint8 f9x16[FONT_9X16_SIZE] = {
//    0xFF, 0x7F, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 
//    0xFE, 0x7F, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 
//    0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 
//    0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 
//    0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 
//    0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 
//    0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 
//    0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 
//    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
//    0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 
//    0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 
//    0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 
//    0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 
//    0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 
//    0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 
//    0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 
//    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
//    0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 
//    0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 
//    0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 
//    0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 
//    0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 
//    0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 
//    0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 
//    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
//    0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 
//    0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 
//    0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 
//    0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 
//    0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 
//    0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x0C, 
//    0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 
//    0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0D, 0xFC, 0x0D, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x01, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x01, 0x00, 0x00, 0x00, 0x00, 0x70, 0x04, 0xF8, 0x0C, 0x88, 0x08, 
//    0xFE, 0x3F, 0x88, 0x08, 0x98, 0x0F, 0x10, 0x07, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x1C, 0x0C, 
//    0x14, 0x0F, 0xC8, 0x03, 0xF0, 0x04, 0x3C, 0x0A, 0x0C, 0x0E, 0x00, 0x04, 0x00, 0x00, 0x80, 0x07, 
//    0xD8, 0x0F, 0x7C, 0x08, 0xE4, 0x0C, 0xBC, 0x07, 0xD8, 0x0F, 0x40, 0x08, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x03, 0xF8, 0x07, 0x0C, 0x0C, 0x04, 0x08, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0x0C, 0x0C, 0xF8, 0x07, 0xF0, 0x03, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0xA0, 0x02, 0xE0, 0x03, 0xC0, 0x01, 0xE0, 0x03, 
//    0xA0, 0x02, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0xE0, 0x03, 
//    0xE0, 0x03, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 
//    0x00, 0x1C, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 
//    0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x0C, 0x00, 0x0F, 0xC0, 0x03, 0xF0, 0x00, 0x3C, 0x00, 0x0C, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0xF8, 0x07, 0xFC, 0x0F, 0x84, 0x09, 0xC4, 0x08, 0x64, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x08, 0x18, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 
//    0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x18, 0x0C, 0x1C, 0x0E, 0x04, 0x0B, 0x84, 0x09, 0xC4, 0x08, 
//    0x7C, 0x08, 0x38, 0x08, 0x00, 0x00, 0x00, 0x00, 0x18, 0x06, 0x1C, 0x0E, 0x44, 0x08, 0x44, 0x08, 
//    0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0xC0, 0x01, 0x60, 0x01, 
//    0x30, 0x01, 0x18, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x7C, 0x04, 0x7C, 0x0C, 
//    0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xC4, 0x0F, 0x84, 0x07, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x07, 
//    0xF8, 0x0F, 0x4C, 0x08, 0x44, 0x08, 0x44, 0x08, 0xC4, 0x0F, 0x80, 0x07, 0x00, 0x00, 0x00, 0x00, 
//    0x04, 0x00, 0x04, 0x00, 0x04, 0x0E, 0x84, 0x0F, 0xE4, 0x01, 0x7C, 0x00, 0x1C, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0xB8, 0x07, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0xFC, 0x08, 0x84, 0x08, 0x84, 0x08, 0x84, 0x0C, 0xFC, 0x07, 
//    0xF8, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x0C, 0x60, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x60, 0x1C, 
//    0x60, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0xC0, 0x01, 
//    0x60, 0x03, 0x30, 0x06, 0x18, 0x0C, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x20, 0x01, 0x20, 0x01, 
//    0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x20, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x08, 0x08, 0x18, 0x0C, 0x30, 0x06, 0x60, 0x03, 0xC0, 0x01, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x38, 0x00, 0x3C, 0x00, 0x04, 0x00, 0x84, 0x0D, 0xC4, 0x0D, 0x7C, 0x00, 0x38, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0xE4, 0x09, 0x14, 0x0A, 0xFC, 0x0B, 0xF8, 0x0B, 
//    0x00, 0x00, 0x00, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x0F, 
//    0xF8, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 
//    0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 
//    0x04, 0x08, 0x1C, 0x0E, 0x18, 0x06, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x08, 
//    0x04, 0x08, 0x0C, 0x0C, 0xF8, 0x07, 0xF0, 0x03, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 
//    0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x44, 0x00, 0x44, 0x00, 0x44, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x84, 0x08, 0x84, 0x08, 0x9C, 0x0F, 0x98, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x08, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x0E, 0x00, 0x08, 0x04, 0x08, 0xFC, 0x0F, 
//    0xFC, 0x07, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0xC0, 0x00, 0xE0, 0x01, 
//    0x30, 0x03, 0x1C, 0x0E, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 
//    0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xF8, 0x0F, 
//    0x30, 0x00, 0x60, 0x00, 0x30, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x60, 0x00, 0xC0, 0x00, 0x80, 0x01, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x00, 0x78, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x0C, 0x04, 0x0C, 0xFC, 0x1F, 
//    0xF8, 0x17, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x01, 0x84, 0x03, 0x84, 0x06, 
//    0xFC, 0x0C, 0x78, 0x08, 0x00, 0x00, 0x00, 0x00, 0x38, 0x06, 0x7C, 0x0E, 0x44, 0x08, 0x44, 0x08, 
//    0x44, 0x08, 0xCC, 0x0F, 0x88, 0x07, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0xFC, 0x07, 0xFC, 0x0F, 
//    0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x00, 0x00, 0x00, 0x7C, 0x00, 
//    0xFC, 0x03, 0x80, 0x0F, 0x00, 0x0C, 0x80, 0x0F, 0xFC, 0x03, 0x7C, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x03, 0x80, 0x01, 0x00, 0x03, 0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x00, 
//    0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x3C, 0x00, 0x70, 0x00, 0xC0, 0x0F, 0xC0, 0x0F, 0x70, 0x00, 
//    0x3C, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x04, 0x0E, 0x04, 0x0F, 0x84, 0x09, 0xC4, 0x08, 0x64, 0x08, 
//    0x3C, 0x08, 0x1C, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 
//    0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x3C, 0x00, 
//    0xF0, 0x00, 0xC0, 0x03, 0x00, 0x0F, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x08, 0x00, 0x0C, 0x00, 0x06, 0x00, 0x06, 0x00, 0x0C, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x03, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0xA0, 0x0F, 0xA0, 0x08, 0xA0, 0x08, 0xA0, 0x08, 0xE0, 0x0F, 
//    0xC0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 
//    0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 
//    0x20, 0x08, 0x60, 0x0C, 0x40, 0x04, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 
//    0x20, 0x08, 0x20, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 
//    0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 
//    0x20, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x24, 0x00, 0x24, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xC0, 0x07, 0xE0, 0x4F, 0x20, 0x48, 0x20, 0x48, 0x20, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 
//    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x08, 0xEC, 0x0F, 0xEC, 0x0F, 0x00, 0x08, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x70, 0x00, 0x40, 0x20, 0x40, 
//    0xEC, 0x7F, 0xEC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x01, 0x80, 0x03, 
//    0xC0, 0x06, 0x60, 0x0C, 0x20, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 
//    0x20, 0x00, 0xE0, 0x0F, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 
//    0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xC0, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xE0, 0x7F, 0xE0, 0x7F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x7F, 
//    0xE0, 0x7F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0xC0, 0x00, 0x60, 0x00, 0x20, 0x00, 
//    0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x08, 0xE0, 0x09, 0x20, 0x09, 0x20, 0x09, 
//    0x20, 0x09, 0x20, 0x0F, 0x20, 0x06, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0xFC, 0x07, 
//    0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x0F, 
//    0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x00, 
//    0xE0, 0x03, 0x00, 0x0F, 0x00, 0x0C, 0x00, 0x0F, 0xE0, 0x03, 0xE0, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x80, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0x60, 0x0C, 0xE0, 0x0E, 0x80, 0x03, 0x00, 0x01, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x4F, 0x00, 0x48, 0x00, 0x48, 0x00, 0x48, 0xE0, 0x7F, 
//    0xE0, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x20, 0x0C, 0x20, 0x0E, 0x20, 0x0B, 0xA0, 0x09, 0xE0, 0x08, 
//    0x60, 0x08, 0x20, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0xF8, 0x07, 0xBC, 0x0F, 
//    0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 
//    0x04, 0x08, 0xBC, 0x0F, 0xF8, 0x07, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 
//    0x0E, 0x00, 0x02, 0x00, 0x06, 0x00, 0x0C, 0x00, 0x08, 0x00, 0x0E, 0x00, 0x06, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
//    0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xFF, 0x80, 0xFF, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFF, 
//    0x80, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
//    0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 
//    0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 
//    0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0xAA, 0xAA, 0x55, 0x55, 0x00, 0x00, 0xAA, 0xAA, 0x00, 0x00, 0x55, 0x55, 0x00, 0x00, 0xAA, 0xAA, 
//    0x00, 0x00, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 
//    0x55, 0x55, 0xAA, 0xAA, 0xFF, 0xFF, 0xFF, 0xFF, 0x55, 0x55, 0xFF, 0xFF, 0xAA, 0xAA, 0xFF, 0xFF, 
//    0x55, 0x55, 0xFF, 0xFF, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 
//    0xFC, 0xFF, 0x04, 0x00, 0x1C, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x07, 
//    0xE0, 0x07, 0xE0, 0x07, 0xE0, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x80, 0x01, 0xC0, 0x03, 0xC0, 0x03, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x01, 
//    0xC0, 0x03, 0x00, 0x06, 0x00, 0x0C, 0xFE, 0x0F, 0xFE, 0x0F, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x40, 0x02, 0x60, 0x03, 0x20, 0x01, 0x60, 0x03, 0x40, 0x02, 0x60, 0x03, 0x20, 0x01, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x40, 0x08, 0xE0, 0x08, 0xB0, 0x09, 0x18, 0x0B, 0x08, 0x0A, 0x00, 0x08, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x0A, 0x18, 0x0B, 0xB0, 0x09, 0xE0, 0x08, 
//    0x40, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x0E, 0x00, 0x08, 0xFF, 0x0F, 
//    0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x1E, 0x00, 
//    0x12, 0x00, 0x1E, 0x00, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x24, 0x00, 
//    0x36, 0x00, 0x3A, 0x00, 0x2E, 0x00, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0xB0, 0x06, 0xB0, 0x06, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0xE0, 0xFF, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x00, 0x00, 0xC0, 0x07, 0xEC, 0x0F, 0x2C, 0x09, 0x20, 0x09, 
//    0x2C, 0x09, 0xEC, 0x09, 0xC0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xFF, 0x80, 0xFF, 
//    0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 
//    0xE0, 0xFF, 0x60, 0x00, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0xE0, 0xFF, 0xE0, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 
//    0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x00, 0x00, 0x00, 0x00, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x00, 0xE0, 0xFF, 0xE0, 0xFF, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x03, 0xFF, 0x03, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 
//    0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x03, 0xFF, 0x03, 0x00, 0x03, 
//    0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
//    0xFF, 0x03, 0xFF, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x01, 
//    0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x60, 0x03, 0x60, 0x03, 
//    0x7F, 0x03, 0x7F, 0x03, 0x00, 0x03, 0xFF, 0x03, 0xFF, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
//    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 
//    0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x7F, 0xFF, 0x7F, 0xFF, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x0F, 0xFB, 0x0F, 0x8B, 0x08, 0x88, 0x08, 
//    0x8B, 0x08, 0x0B, 0x08, 0x08, 0x08, 0x00, 0x00, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 
//    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x60, 0x03, 0x60, 0x03, 0x7F, 0xFF, 
//    0x7F, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x60, 0x03, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x80, 0x01, 
//    0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0xFF, 0x80, 0xFF, 0x80, 0x01, 0x80, 0x01, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x00, 0x60, 0xFF, 0x60, 0xFF, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 
//    0x60, 0x03, 0x60, 0x03, 0x80, 0x01, 0x80, 0x01, 0xFF, 0x01, 0xFF, 0x01, 0x80, 0x01, 0xFF, 0x01, 
//    0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x60, 0x03, 0x60, 0x03, 0x7F, 0x03, 0x7F, 0x03, 0x00, 0x03, 
//    0x7F, 0x03, 0x7F, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 
//    0xFF, 0xFF, 0xFF, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x60, 0x03, 0x80, 0x01, 0x80, 0x01, 0xFF, 0xFF, 
//    0xFF, 0xFF, 0x80, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x01, 0x80, 0x01, 0x60, 0x03, 0x60, 0x03, 
//    0x7F, 0xFF, 0x7F, 0xFF, 0x00, 0x00, 0x7F, 0xFF, 0x7F, 0xFF, 0x60, 0x03, 0x60, 0x03, 0x00, 0x00, 
//    0xF0, 0x07, 0x08, 0x08, 0xC8, 0x09, 0x28, 0x0A, 0x28, 0x0A, 0x48, 0x09, 0x08, 0x08, 0xF0, 0x07, 
//    0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x01, 0xC0, 0x07, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0xA0, 0x0F, 0xA0, 0x08, 0xA0, 0x08, 0xA0, 0x08, 0xE0, 0x0F, 
//    0xC0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x24, 0x08, 0x24, 0x08, 0x24, 0x08, 
//    0xE4, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x00, 0x08, 
//    0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x3F, 0x00, 0x38, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x4F, 0x20, 0x48, 
//    0x20, 0x48, 0x20, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 
//    0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x01, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 
//    0xE0, 0x0F, 0x20, 0x08, 0xF0, 0x1F, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 
//    0x00, 0x06, 0x20, 0x0F, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 0xE0, 0x09, 0xC0, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x60, 0x0C, 0xE0, 0x0E, 0x80, 0x03, 0x00, 0x01, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0xE0, 0x0F, 
//    0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE4, 0x0F, 0x0C, 0x08, 0x08, 0x08, 0x0C, 0x08, 
//    0xE4, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x01, 0x80, 0x03, 
//    0xC0, 0x06, 0x60, 0x0C, 0x20, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0xC0, 0x0F, 0xE0, 0x07, 
//    0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 
//    0xC0, 0x00, 0x80, 0x01, 0xC0, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 
//    0xE0, 0x0F, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 
//    0x00, 0x00, 0x00, 0x00, 0xC0, 0x08, 0xE0, 0x0D, 0x20, 0x07, 0x20, 0x03, 0x20, 0x01, 0xE0, 0x0F, 
//    0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x7F, 0xE0, 0x7F, 0x20, 0x08, 0x20, 0x08, 0x20, 0x08, 
//    0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0x0F, 0x20, 0x08, 0x20, 0x08, 
//    0x20, 0x08, 0x60, 0x0C, 0x40, 0x04, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 
//    0xE0, 0x0F, 0xE0, 0x0F, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x4F, 
//    0x00, 0x48, 0x00, 0x48, 0x00, 0x48, 0xE0, 0x7F, 0xE0, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x60, 0x0C, 
//    0xE0, 0x0E, 0x80, 0x03, 0xE0, 0x0F, 0x80, 0x03, 0xE0, 0x0E, 0x60, 0x0C, 0x00, 0x00, 0x00, 0x00, 
//    0xF8, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x7C, 0x08, 0xD8, 0x0F, 0x80, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x80, 0x08, 0x80, 0x0F, 0x00, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x00, 0x07, 0x00, 0x00, 0xE0, 0x0F, 
//    0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x40, 0x04, 0x60, 0x0C, 0x20, 0x09, 0x20, 0x09, 0x20, 0x09, 
//    0xE0, 0x0F, 0xC0, 0x06, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 
//    0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x40, 0x04, 0x60, 0x0C, 0x20, 0x09, 
//    0x20, 0x09, 0x20, 0x09, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0xE0, 0x0F, 
//    0x00, 0x08, 0xE0, 0x0F, 0x00, 0x08, 0xE0, 0x0F, 0xE0, 0x3F, 0x00, 0x38, 0x00, 0x00, 0xE0, 0x00, 
//    0xE0, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xE0, 0x0F, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0x20, 0x00, 0xE0, 0x0F, 0xE0, 0x0F, 0x80, 0x08, 0x80, 0x08, 0x80, 0x0F, 0x00, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x80, 0x00, 0xF8, 0x07, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0xFC, 0x0F, 
//    0xF8, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x24, 0x08, 0x24, 0x08, 0x24, 0x08, 
//    0xE4, 0x0F, 0xC0, 0x07, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0x00, 0x08, 
//    0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x3F, 0x00, 0x38, 0x00, 0x00, 0x00, 0x18, 0xF8, 0x1F, 0xFC, 0x0F, 
//    0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xFC, 0x1F, 0x00, 0x18, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 
//    0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0x04, 0x08, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x07, 
//    0xFC, 0x0F, 0x04, 0x08, 0xFE, 0x1F, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 0x00, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 
//    0x00, 0x00, 0x0C, 0x0C, 0x3C, 0x0F, 0xF0, 0x03, 0xC0, 0x00, 0xF0, 0x03, 0x3C, 0x0F, 0x0C, 0x0C, 
//    0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x80, 0x01, 0xC0, 0x00, 0x60, 0x00, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFD, 0x0F, 0x83, 0x01, 0xC2, 0x00, 0x63, 0x00, 
//    0xFD, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0xC0, 0x00, 0xE0, 0x01, 
//    0x30, 0x03, 0x1C, 0x0E, 0x0C, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0xF0, 0x0F, 0xF8, 0x07, 
//    0x0C, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xF8, 0x0F, 
//    0x30, 0x00, 0x60, 0x00, 0x30, 0x00, 0xF8, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 
//    0x00, 0x00, 0x00, 0x00, 0x78, 0x08, 0xFC, 0x0C, 0x84, 0x06, 0x84, 0x03, 0x84, 0x01, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 
//    0xFC, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x07, 0xFC, 0x0F, 0x04, 0x08, 0x04, 0x08, 
//    0x04, 0x08, 0x1C, 0x0E, 0x18, 0x06, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x7C, 0x00, 0xFC, 0x08, 
//    0x80, 0x08, 0x80, 0x08, 0x80, 0x08, 0xFC, 0x0F, 0xFC, 0x07, 0x00, 0x00, 0x00, 0x00, 0x1C, 0x0F, 
//    0xBC, 0x0F, 0xE0, 0x00, 0xFC, 0x0F, 0xE0, 0x00, 0xBC, 0x0F, 0x1C, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0xFC, 0x0F, 0xFC, 0x0F, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 
//    0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 
//    0x00, 0x00, 0x00, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0xC0, 0x07, 0x00, 0x00, 0xFC, 0x0F, 
//    0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x18, 0x06, 0x1C, 0x0E, 0x44, 0x08, 0x44, 0x08, 0x44, 0x08, 
//    0xFC, 0x0F, 0xB8, 0x07, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 
//    0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x08, 0x04, 0x0C, 0x0C, 0x44, 0x08, 
//    0x44, 0x08, 0x44, 0x08, 0xFC, 0x0F, 0xF8, 0x07, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0xFC, 0x0F, 
//    0x00, 0x08, 0xFC, 0x0F, 0x00, 0x08, 0xFC, 0x0F, 0xFC, 0x3F, 0x00, 0x38, 0x00, 0x00, 0x7C, 0x00, 
//    0xFC, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x00, 0x00, 0x00, 0x00, 
//    0x04, 0x00, 0xFC, 0x0F, 0xFC, 0x0F, 0x20, 0x08, 0x20, 0x08, 0xE0, 0x0F, 0xC0, 0x07, 0x00, 0x00, 
//};

const uint8 f15x24[FONT_15X24_SIZE] = {
    /*0000*/    0xFF, 0xFF, 0x7F, 0x01, 0x00, 0x80, 0x19, 0x00, 0x98, 0x79, 0x00, 0x9E, 0xE1, 0x81, 0x87, 0x81, 
    /*0010*/    0xE7, 0x81, 0x01, 0x7E, 0x80, 0x01, 0x18, 0x80, 0x01, 0x7E, 0x80, 0x81, 0xE7, 0x81, 0xE1, 0x81, 
    /*0020*/    0x87, 0x79, 0x00, 0x9E, 0x19, 0x00, 0x98, 0x01, 0x00, 0x80, 0xFE, 0xFF, 0xFF, 0x00, 0x00, 0x00, 
    /*0030*/    0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 
    /*0040*/    0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 
    /*0050*/    0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0060*/    0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 
    /*0070*/    0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 
    /*0080*/    0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 
    /*0090*/    0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 
    /*00A0*/    0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 
    /*00B0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 
    /*00C0*/    0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 
    /*00D0*/    0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*00E0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 
    /*00F0*/    0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 
    /*0100*/    0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0110*/    0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 
    /*0120*/    0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 
    /*0130*/    0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0140*/    0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 
    /*0150*/    0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 
    /*0160*/    0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 
    /*0170*/    0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 
    /*0180*/    0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 
    /*0190*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 
    /*01A0*/    0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 
    /*01B0*/    0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 
    /*01C0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 
    /*01D0*/    0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 
    /*01E0*/    0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*01F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 
    /*0200*/    0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 
    /*0210*/    0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0220*/    0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 
    /*0230*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 
    /*0240*/    0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 
    /*0250*/    0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 
    /*0260*/    0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 
    /*0270*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 
    /*0280*/    0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 
    /*0290*/    0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 
    /*02A0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x78, 0x00, 0x1E, 0xE0, 
    /*02B0*/    0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x7E, 0x00, 0x80, 0xE7, 
    /*02C0*/    0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*02D0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*02E0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*02F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0300*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0310*/    0x00, 0x00, 0xF8, 0x7F, 0x1E, 0xF8, 0x7F, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0320*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0330*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0340*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0350*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x81, 0x01, 
    /*0360*/    0x80, 0x81, 0x01, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x80, 0x81, 0x01, 0x80, 0x81, 0x01, 0x80, 
    /*0370*/    0x81, 0x01, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x80, 0x81, 0x01, 0x80, 0x81, 0x01, 0x00, 0x00, 
    /*0380*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0x03, 
    /*0390*/    0xE0, 0x0F, 0x07, 0x30, 0x18, 0x0C, 0x30, 0x18, 0x0C, 0xFC, 0xFF, 0x3F, 0xFC, 0xFF, 0x3F, 0x30, 
    /*03A0*/    0x18, 0x0C, 0x30, 0x18, 0x0C, 0xE0, 0xF0, 0x07, 0xC0, 0xE0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*03B0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0x00, 0x00, 0xF8, 0x00, 0x18, 0xD8, 0x00, 0x1E, 
    /*03C0*/    0xD8, 0x80, 0x07, 0xF8, 0xE0, 0x01, 0x70, 0x78, 0x00, 0x00, 0x1E, 0x0E, 0x80, 0x07, 0x1F, 0xE0, 
    /*03D0*/    0x01, 0x1B, 0x78, 0x00, 0x1B, 0x18, 0x00, 0x1F, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*03E0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0xE0, 0xE1, 0x0F, 0xF0, 0x73, 0x1C, 0x18, 0x3F, 0x18, 
    /*03F0*/    0x18, 0x1E, 0x18, 0x18, 0x3E, 0x18, 0x18, 0x77, 0x1C, 0xF0, 0xE3, 0x0F, 0xE0, 0xC1, 0x07, 0x00, 
    /*0400*/    0xE0, 0x0F, 0x00, 0x70, 0x1C, 0x00, 0x30, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0410*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0420*/    0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0430*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0440*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xFF, 0x01, 0xE0, 0xFF, 0x07, 0x70, 0x00, 0x0E, 
    /*0450*/    0x18, 0x00, 0x18, 0x08, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0460*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0470*/    0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x10, 0x18, 0x00, 0x18, 0x70, 0x00, 0x0E, 0xE0, 0xFF, 0x07, 
    /*0480*/    0x80, 0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0490*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x80, 0x18, 0x01, 0x80, 0x99, 
    /*04A0*/    0x01, 0x80, 0xDB, 0x01, 0x00, 0xFF, 0x00, 0x00, 0x7E, 0x00, 0x00, 0xFF, 0x00, 0x80, 0xDB, 0x01, 
    /*04B0*/    0x80, 0x99, 0x01, 0x80, 0x18, 0x01, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*04C0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*04D0*/    0x00, 0x00, 0x18, 0x00, 0xC0, 0xFF, 0x03, 0xC0, 0xFF, 0x03, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*04E0*/    0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*04F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 
    /*0500*/    0x60, 0x00, 0x00, 0x3E, 0x00, 0x00, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0510*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0520*/    0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*0530*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*0540*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0550*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1E, 0x00, 0x00, 
    /*0560*/    0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0570*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 
    /*0580*/    0x00, 0x1E, 0x00, 0x80, 0x07, 0x00, 0xE0, 0x01, 0x00, 0x78, 0x00, 0x00, 0x1E, 0x00, 0x80, 0x07, 
    /*0590*/    0x00, 0xE0, 0x01, 0x00, 0x78, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*05A0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 0xE0, 0x18, 0x18, 
    /*05B0*/    0x70, 0x18, 0x18, 0x38, 0x18, 0x18, 0x1C, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x07, 0x18, 0x98, 0x03, 
    /*05C0*/    0x18, 0xF0, 0xFF, 0x0F, 0xE0, 0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*05D0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x18, 0x60, 0x00, 0x18, 0x70, 
    /*05E0*/    0x00, 0x18, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 
    /*05F0*/    0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0600*/    0xE0, 0x01, 0x1C, 0xF0, 0x01, 0x1E, 0x18, 0x00, 0x1F, 0x18, 0x80, 0x1B, 0x18, 0xC0, 0x19, 0x18, 
    /*0610*/    0xE0, 0x18, 0x18, 0x70, 0x18, 0x18, 0x38, 0x18, 0x18, 0x1C, 0x18, 0xF0, 0x0F, 0x18, 0xE0, 0x07, 
    /*0620*/    0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x06, 
    /*0630*/    0x70, 0x00, 0x0E, 0x18, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 
    /*0640*/    0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF0, 0xFF, 0x0F, 0xE0, 0xE7, 0x07, 0x00, 0x00, 
    /*0650*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x01, 0x00, 0xF8, 0x01, 
    /*0660*/    0x00, 0x9C, 0x01, 0x00, 0x8E, 0x01, 0x00, 0x87, 0x01, 0x80, 0x83, 0x01, 0xC0, 0x81, 0x01, 0xE0, 
    /*0670*/    0x80, 0x01, 0x70, 0x80, 0x01, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0680*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x0F, 0x06, 0xF8, 0x0F, 0x0E, 0x18, 0x0C, 0x18, 
    /*0690*/    0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 
    /*06A0*/    0x0C, 0x18, 0x18, 0xF8, 0x0F, 0x18, 0xF0, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*06B0*/    0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 
    /*06C0*/    0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 
    /*06D0*/    0xF8, 0x0F, 0x00, 0xF0, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*06E0*/    0x00, 0xF8, 0x00, 0x00, 0xF8, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 
    /*06F0*/    0x18, 0xC0, 0x1F, 0x18, 0xF0, 0x1F, 0x18, 0x3C, 0x00, 0x18, 0x0F, 0x00, 0xF8, 0x03, 0x00, 0xF8, 
    /*0700*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xE7, 
    /*0710*/    0x07, 0xF0, 0xFF, 0x0F, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 
    /*0720*/    0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF0, 0xFF, 0x0F, 0xE0, 0xE7, 0x07, 0x00, 
    /*0730*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x0F, 0x00, 0xF0, 0x1F, 
    /*0740*/    0x18, 0x18, 0x30, 0x18, 0x18, 0x30, 0x18, 0x18, 0x30, 0x18, 0x18, 0x30, 0x18, 0x18, 0x30, 0x18, 
    /*0750*/    0x18, 0x30, 0x18, 0x18, 0x30, 0x18, 0xF0, 0xFF, 0x0F, 0xE0, 0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 
    /*0760*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0770*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x07, 0x1E, 0x80, 0x07, 0x1E, 0x00, 0x00, 0x00, 
    /*0780*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0790*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07A0*/    0x40, 0x00, 0x00, 0x60, 0x80, 0x07, 0x3E, 0x80, 0x07, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07B0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07C0*/    0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x38, 0x00, 0x00, 0x7C, 0x00, 0x00, 0xEE, 0x00, 0x00, 0xC7, 
    /*07D0*/    0x01, 0x80, 0x83, 0x03, 0xC0, 0x01, 0x07, 0xE0, 0x00, 0x0E, 0x70, 0x00, 0x1C, 0x30, 0x00, 0x18, 
    /*07E0*/    0x10, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*07F0*/    0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 
    /*0800*/    0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 0x00, 0xC6, 0x00, 
    /*0810*/    0x00, 0xC6, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x30, 
    /*0820*/    0x00, 0x18, 0x70, 0x00, 0x1C, 0xE0, 0x00, 0x0E, 0xC0, 0x01, 0x07, 0x80, 0x83, 0x03, 0x00, 0xC7, 
    /*0830*/    0x01, 0x00, 0xEE, 0x00, 0x00, 0x7C, 0x00, 0x00, 0x38, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 
    /*0840*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x01, 0x00, 0xF0, 0x01, 0x00, 0x18, 
    /*0850*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x70, 0x1E, 0x18, 0x78, 0x1E, 0x18, 0x0C, 
    /*0860*/    0x00, 0x18, 0x06, 0x00, 0xF0, 0x03, 0x00, 0xE0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0870*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 0x00, 0x18, 0x18, 
    /*0880*/    0x00, 0x18, 0x18, 0x7E, 0x18, 0x18, 0xFF, 0x18, 0x98, 0x81, 0x19, 0x98, 0x81, 0x19, 0x98, 0x81, 
    /*0890*/    0x19, 0x98, 0x81, 0x19, 0xF0, 0xFF, 0x18, 0xE0, 0xFF, 0x19, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*08A0*/    0x00, 0x00, 0x00, 0xE0, 0xFF, 0x1F, 0xF0, 0xFF, 0x1F, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 
    /*08B0*/    0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0xF0, 0xFF, 
    /*08C0*/    0x1F, 0xE0, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*08D0*/    0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 
    /*08E0*/    0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x1E, 0x18, 0xF0, 0xF3, 0x0F, 0xE0, 0xE1, 
    /*08F0*/    0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 
    /*0900*/    0xF0, 0xFF, 0x0F, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 
    /*0910*/    0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0xF0, 0x00, 0x0F, 0xE0, 0x00, 0x07, 0x00, 0x00, 
    /*0920*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 
    /*0930*/    0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x38, 
    /*0940*/    0x00, 0x1C, 0x70, 0x00, 0x0E, 0xE0, 0xFF, 0x07, 0xC0, 0xFF, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0950*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x18, 0x18, 
    /*0960*/    0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 
    /*0970*/    0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0980*/    0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 
    /*0990*/    0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*09A0*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*09B0*/    0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 
    /*09C0*/    0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xF0, 0xF8, 0x0F, 0xE0, 
    /*09D0*/    0xF8, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 
    /*09E0*/    0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*09F0*/    0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 
    /*0A00*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0A10*/    0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 
    /*0A20*/    0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0A30*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x07, 0x00, 0x80, 0x0F, 0x00, 0x00, 
    /*0A40*/    0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 
    /*0A50*/    0xF8, 0xFF, 0x0F, 0xF8, 0xFF, 0x07, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0A60*/    0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x3C, 0x00, 0x00, 0x7E, 
    /*0A70*/    0x00, 0x00, 0xE7, 0x00, 0x80, 0xC3, 0x01, 0xC0, 0x81, 0x03, 0xE0, 0x00, 0x07, 0x70, 0x00, 0x0E, 
    /*0A80*/    0x38, 0x00, 0x1C, 0x18, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0A90*/    0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 
    /*0AA0*/    0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*0AB0*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 
    /*0AC0*/    0xFF, 0x1F, 0xF0, 0xFF, 0x1F, 0xE0, 0x00, 0x00, 0xC0, 0x01, 0x00, 0x80, 0x03, 0x00, 0x00, 0x07, 
    /*0AD0*/    0x00, 0x00, 0x07, 0x00, 0x80, 0x03, 0x00, 0xC0, 0x01, 0x00, 0xE0, 0x00, 0x00, 0xF0, 0xFF, 0x1F, 
    /*0AE0*/    0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 
    /*0AF0*/    0xFF, 0x1F, 0x80, 0x03, 0x00, 0x00, 0x07, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1C, 0x00, 0x00, 0x38, 
    /*0B00*/    0x00, 0x00, 0x70, 0x00, 0x00, 0xE0, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 
    /*0B10*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 
    /*0B20*/    0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 
    /*0B30*/    0x18, 0x18, 0x00, 0x18, 0xF0, 0xFF, 0x0F, 0xE0, 0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B40*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x30, 0x00, 0x18, 
    /*0B50*/    0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 0x00, 0x18, 0x30, 
    /*0B60*/    0x00, 0xF0, 0x1F, 0x00, 0xE0, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0B70*/    0x00, 0x00, 0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xFF, 0x0F, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 
    /*0B80*/    0x00, 0x18, 0x18, 0x00, 0x1A, 0x18, 0x00, 0x1E, 0x18, 0x00, 0x1E, 0x18, 0x00, 0x3C, 0xF0, 0xFF, 
    /*0B90*/    0x7F, 0xE0, 0xFF, 0x67, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0BA0*/    0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x70, 0x00, 0x18, 0xF0, 0x00, 0x18, 0xF0, 0x01, 0x18, 
    /*0BB0*/    0xB0, 0x03, 0x18, 0x30, 0x07, 0x18, 0x30, 0x0E, 0x18, 0x30, 0x1C, 0xF0, 0x1F, 0x18, 0xE0, 0x0F, 
    /*0BC0*/    0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0x07, 0x06, 
    /*0BD0*/    0xF0, 0x0F, 0x0E, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 
    /*0BE0*/    0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x70, 0xF0, 0x0F, 0x60, 0xE0, 0x07, 0x00, 0x00, 
    /*0BF0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 
    /*0C00*/    0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 
    /*0C10*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 
    /*0C20*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x07, 0xF8, 0xFF, 0x0F, 0x00, 0x00, 0x18, 
    /*0C30*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*0C40*/    0x00, 0x18, 0xF8, 0xFF, 0x0F, 0xF8, 0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0C50*/    0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0xF8, 0x07, 0x00, 0x80, 0x7F, 0x00, 0x00, 0xF8, 0x03, 
    /*0C60*/    0x00, 0x80, 0x1F, 0x00, 0x00, 0x1C, 0x00, 0x80, 0x1F, 0x00, 0xF8, 0x03, 0x80, 0x7F, 0x00, 0xF8, 
    /*0C70*/    0x07, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0C80*/    0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x0F, 0x00, 0x00, 0x07, 0x00, 0x80, 0x03, 0x00, 0xC0, 0x01, 
    /*0C90*/    0x00, 0xE0, 0x00, 0x00, 0xE0, 0x00, 0x00, 0xC0, 0x01, 0x00, 0x80, 0x03, 0x00, 0x00, 0x07, 0xF8, 
    /*0CA0*/    0xFF, 0x0F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 
    /*0CB0*/    0x18, 0x78, 0x00, 0x1E, 0xE0, 0x81, 0x07, 0x80, 0xE7, 0x01, 0x00, 0x7E, 0x00, 0x00, 0x18, 0x00, 
    /*0CC0*/    0x00, 0x7E, 0x00, 0x80, 0xE7, 0x01, 0xE0, 0x81, 0x07, 0x78, 0x00, 0x1E, 0x18, 0x00, 0x18, 0x00, 
    /*0CD0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x78, 0x00, 
    /*0CE0*/    0x00, 0xE0, 0x01, 0x00, 0x80, 0x07, 0x00, 0x00, 0x1E, 0x00, 0x00, 0xF8, 0x1F, 0x00, 0xF8, 0x1F, 
    /*0CF0*/    0x00, 0x1E, 0x00, 0x80, 0x07, 0x00, 0xE0, 0x01, 0x00, 0x78, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 
    /*0D00*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x80, 0x1F, 0x18, 0xC0, 0x1F, 0x18, 0xE0, 
    /*0D10*/    0x18, 0x18, 0x70, 0x18, 0x18, 0x38, 0x18, 0x18, 0x1C, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x07, 0x18, 
    /*0D20*/    0x98, 0x03, 0x18, 0xF8, 0x01, 0x18, 0xF8, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0D30*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 
    /*0D40*/    0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 
    /*0D50*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0D60*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x78, 0x00, 0x00, 0xE0, 0x01, 0x00, 0x80, 0x07, 
    /*0D70*/    0x00, 0x00, 0x1E, 0x00, 0x00, 0x78, 0x00, 0x00, 0xE0, 0x01, 0x00, 0x80, 0x07, 0x00, 0x00, 0x1E, 
    /*0D80*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0D90*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 
    /*0DA0*/    0x18, 0x18, 0x00, 0x18, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0DB0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 
    /*0DC0*/    0x00, 0x00, 0x30, 0x00, 0x00, 0x38, 0x00, 0x00, 0x1C, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x0E, 0x00, 
    /*0DD0*/    0x00, 0x1C, 0x00, 0x00, 0x38, 0x00, 0x00, 0x30, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0DE0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 
    /*0DF0*/    0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 
    /*0E00*/    0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0E10*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 
    /*0E20*/    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0E30*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0E40*/    0x00, 0x00, 0x00, 0x00, 0xC0, 0x07, 0x00, 0xE0, 0x0F, 0x00, 0x33, 0x18, 0x00, 0x33, 0x18, 0x00, 
    /*0E50*/    0x33, 0x18, 0x00, 0x33, 0x18, 0x00, 0x33, 0x18, 0x00, 0x33, 0x18, 0x00, 0x33, 0x18, 0x00, 0xFE, 
    /*0E60*/    0x1F, 0x00, 0xFC, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0E70*/    0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 
    /*0E80*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0xFE, 0x0F, 0x00, 0xFC, 
    /*0E90*/    0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 
    /*0EA0*/    0x00, 0xFE, 0x0F, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 
    /*0EB0*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x07, 0x1C, 0x00, 0x06, 0x0C, 0x00, 0x04, 0x04, 0x00, 0x00, 
    /*0EC0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0x00, 0xFE, 0x0F, 
    /*0ED0*/    0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 
    /*0EE0*/    0x03, 0x18, 0x00, 0x03, 0x18, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0EF0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0x00, 0xFE, 0x0F, 0x00, 0x63, 0x18, 
    /*0F00*/    0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 
    /*0F10*/    0x63, 0x1C, 0x00, 0x7E, 0x0C, 0x00, 0x7C, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0F20*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    /*0F30*/    0xF0, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x18, 0x03, 0x00, 0x18, 0x03, 0x00, 0x18, 0x03, 0x00, 0x18, 
    /*0F40*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0F50*/    0x00, 0x00, 0xFC, 0x07, 0x00, 0xFE, 0x0F, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 
    /*0F60*/    0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0xFF, 0xFF, 0x00, 
    /*0F70*/    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 
    /*0F80*/    0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    /*0F90*/    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0xFE, 0x1F, 0x00, 0xFC, 0x1F, 0x00, 
    /*0FA0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0FB0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x38, 0xFF, 0x1F, 0x38, 0xFF, 0x1F, 
    /*0FC0*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*0FD0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 
    /*0FE0*/    0x78, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x00, 0x03, 0x60, 0x00, 0x03, 0x60, 
    /*0FF0*/    0x38, 0xFF, 0x7F, 0x38, 0xFF, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1000*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0xE0, 
    /*1010*/    0x00, 0x00, 0xF0, 0x01, 0x00, 0xB8, 0x03, 0x00, 0x1C, 0x07, 0x00, 0x0E, 0x0E, 0x00, 0x07, 0x1C, 
    /*1020*/    0x00, 0x03, 0x18, 0x00, 0x01, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1030*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x18, 0x00, 
    /*1040*/    0x18, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 
    /*1050*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1060*/    0xFF, 0x1F, 0x00, 0xFF, 0x1F, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0xFF, 
    /*1070*/    0x1F, 0x00, 0xFF, 0x1F, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0xFE, 0x1F, 
    /*1080*/    0x00, 0xFC, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x1F, 0x00, 
    /*1090*/    0xFF, 0x1F, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 
    /*10A0*/    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0xFE, 0x1F, 0x00, 0xFC, 0x1F, 0x00, 0x00, 0x00, 
    /*10B0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0x00, 0xFE, 0x0F, 0x00, 
    /*10C0*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 
    /*10D0*/    0x18, 0x00, 0x03, 0x18, 0x00, 0xFE, 0x0F, 0x00, 0xFC, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*10E0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0x00, 0x03, 0x18, 0x00, 
    /*10F0*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 
    /*1100*/    0x18, 0x00, 0xFE, 0x0F, 0x00, 0xFC, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1110*/    0x00, 0x00, 0x00, 0x00, 0xFC, 0x07, 0x00, 0xFE, 0x0F, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 
    /*1120*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0xFF, 
    /*1130*/    0xFF, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1140*/    0x00, 0xFF, 0x1F, 0x00, 0xFF, 0x1F, 0x00, 0x1C, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x07, 0x00, 0x00, 
    /*1150*/    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 
    /*1160*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1C, 0x04, 
    /*1170*/    0x00, 0x3E, 0x0C, 0x00, 0x63, 0x1C, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 
    /*1180*/    0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0x63, 0x18, 0x00, 0xC6, 0x0F, 0x00, 0x84, 0x07, 0x00, 0x00, 
    /*1190*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 
    /*11A0*/    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0xF8, 0xFF, 0x0F, 0xF8, 0xFF, 0x1F, 0x00, 0x03, 0x18, 0x00, 
    /*11B0*/    0x03, 0x18, 0x00, 0x03, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*11C0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x07, 0x00, 0xFF, 0x0F, 0x00, 0x00, 0x18, 
    /*11D0*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*11E0*/    0x00, 0x18, 0x00, 0xFF, 0x1F, 0x00, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*11F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0x00, 0x00, 0x3F, 0x00, 0x00, 0xF8, 0x01, 0x00, 0xC0, 0x07, 
    /*1200*/    0x00, 0x00, 0x1E, 0x00, 0x00, 0x18, 0x00, 0x00, 0x1E, 0x00, 0xC0, 0x07, 0x00, 0xF8, 0x01, 0x00, 
    /*1210*/    0x3F, 0x00, 0x00, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1220*/    0x00, 0x00, 0xFF, 0x07, 0x00, 0xFF, 0x0F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*1230*/    0x00, 0xF0, 0x1F, 0x00, 0xF0, 0x1F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 
    /*1240*/    0xFF, 0x0F, 0x00, 0xFF, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 
    /*1250*/    0x1C, 0x00, 0x0F, 0x1E, 0x00, 0x1C, 0x07, 0x00, 0xB8, 0x03, 0x00, 0xF0, 0x01, 0x00, 0xE0, 0x00, 
    /*1260*/    0x00, 0xF0, 0x01, 0x00, 0xB8, 0x03, 0x00, 0x1C, 0x07, 0x00, 0x0F, 0x1E, 0x00, 0x07, 0x1C, 0x00, 
    /*1270*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x07, 0x00, 0xFF, 
    /*1280*/    0x0F, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*1290*/    0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 
    /*12A0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x1C, 0x00, 0x03, 0x1E, 0x00, 0x03, 
    /*12B0*/    0x1F, 0x00, 0x83, 0x1B, 0x00, 0xC3, 0x19, 0x00, 0xE3, 0x18, 0x00, 0x73, 0x18, 0x00, 0x3B, 0x18, 
    /*12C0*/    0x00, 0x1F, 0x18, 0x00, 0x0F, 0x18, 0x00, 0x07, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*12D0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 
    /*12E0*/    0x00, 0xE0, 0xFF, 0x07, 0xF0, 0xE7, 0x0F, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 
    /*12F0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1300*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1310*/    0x00, 0xF8, 0xFF, 0x1F, 0xF8, 0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1320*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1330*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x18, 0xF0, 0xE7, 
    /*1340*/    0x0F, 0xE0, 0xFF, 0x07, 0x00, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1350*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3C, 0x00, 0x00, 0x3E, 
    /*1360*/    0x00, 0x00, 0x06, 0x00, 0x00, 0x06, 0x00, 0x00, 0x06, 0x00, 0x00, 0x1C, 0x00, 0x00, 0x38, 0x00, 
    /*1370*/    0x00, 0x30, 0x00, 0x00, 0x30, 0x00, 0x00, 0x3E, 0x00, 0x00, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1380*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*1390*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    /*13A0*/    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
};

